from .array import *  # NOQA
from .interfaces import *  # NOQA
from .line_halfs import *  # NOQA
from .lines import *  # NOQA
from .linestrings import *  # NOQA
from .nodes import *  # NOQA
from .pumps import *  # NOQA
from .quarters import *  # NOQA
from .settings import *  # NOQA
from .surfaces import *  # NOQA



import numpy as np

from threedigrid_builder.base import Array, search
from threedigrid_builder.constants import ContentType

__all__ = ["Windshieldings"]


class Windshielding:
    id: int
    channel_id: int
    north: float
    northeast: float
    east: float
    southeast: float
    south: float
    southwest: float
    west: float
    northwest: float


class Windshieldings(Array[Windshielding]):
    def apply_to_lines(self, lines):
        """Set windshielding factors on lines. Only channels can have windshielding,
        as only channels can have wind applied to them

            Args:
                lines (Lines)
        """
        if len(self) == 0:
            return

        line_ch_idx = np.where(lines.content_type == ContentType.TYPE_V2_CHANNEL)[0]
        has_windshielding = line_ch_idx[
            np.isin(lines.content_pk[line_ch_idx], self.channel_id)
        ]

        chn_idx = search(
            self.channel_id,
            lines.content_pk[has_windshielding],
            check_exists=False,
            assume_ordered=False,
        )

        lines.windshieldings[has_windshielding, 0] = self.north[chn_idx]
        lines.windshieldings[has_windshielding, 1] = self.northeast[chn_idx]
        lines.windshieldings[has_windshielding, 2] = self.east[chn_idx]
        lines.windshieldings[has_windshielding, 3] = self.southeast[chn_idx]
        lines.windshieldings[has_windshielding, 4] = self.south[chn_idx]
        lines.windshieldings[has_windshielding, 5] = self.southwest[chn_idx]
        lines.windshieldings[has_windshielding, 6] = self.west[chn_idx]
        lines.windshieldings[has_windshielding, 7] = self.northwest[chn_idx]



# Module containing functions to read from the SQLite

import pathlib
from contextlib import contextmanager
from enum import Enum
from functools import lru_cache
from typing import Callable, ContextManager

import numpy as np
import shapely
from condenser import NumpyQuery
from pyproj import Transformer
from pyproj.crs import CRS
from sqlalchemy import cast, func, inspect, Integer
from sqlalchemy.orm import Session
from threedi_schema import custom_types, models, ModelSchema, ThreediDatabase

from threedigrid_builder.base import GridSettings, Pumps, TablesSettings
from threedigrid_builder.constants import InitializationType, LineType
from threedigrid_builder.exceptions import SchematisationError
from threedigrid_builder.grid import (
    BoundaryConditions1D,
    BoundaryConditions2D,
    Channels,
    ConnectionNodes,
    CrossSectionDefinitions,
    CrossSectionLocations,
    Culverts,
    DemAverageAreas,
    ExchangeLines,
    GridRefinements,
    Obstacles,
    Orifices,
    Pipes,
    PotentialBreaches,
    Surfaces,
    Weirs,
    Windshieldings,
)

__all__ = ["SQLite"]

# hardcoded source projection
SOURCE_EPSG = 4326

MIN_SQLITE_VERSION = 224

DAY_IN_SECONDS = 24.0 * 3600.0

# put some global defaults on datatypes
NumpyQuery.default_numpy_settings[Integer] = {"dtype": np.int32, "null": -9999}
NumpyQuery.default_numpy_settings[custom_types.IntegerEnum] = {
    **NumpyQuery.default_numpy_settings[Integer],
    "sql_cast": lambda x: cast(x, Integer),
}
NumpyQuery.default_numpy_settings[custom_types.Geometry] = {
    "dtype": np.dtype("O"),
    "sql_cast": func.ST_AsBinary,
    "numpy_cast": shapely.from_wkb,
}


def _object_as_dict(obj) -> dict:
    """Convert SQLAlchemy object to dict, casting Enums. Optional to prefix keys."""
    result = {}
    if obj is None:
        return result
    for c in inspect(obj).mapper.column_attrs:
        val = getattr(obj, c.key)
        if isinstance(val, Enum):
            val = val.value
        result[c.key] = val
    return result


def _set_initialization_type(
    dct, global_field, file_field=None, type_field=None, default=None
):
    """Set the InitializationType depending on global_field and file_field."""
    if not file_field:
        file_field = f"{global_field}_file"
    if not type_field:
        type_field = f"{global_field}_type"
    # If the ``file_field`` contains a value, the initialization type will be changed to
    # the ``default_type``, if supplied.
    if dct[file_field]:
        if default is not None:
            dct[type_field] = default
    # If there is no file, check if the global value is not None
    elif dct[global_field] is not None:
        dct[type_field] = InitializationType.GLOBAL
    else:
        # No file, no global value
        dct[type_field] = None


class SQLite:
    def __init__(self, path: pathlib.Path, upgrade=False, convert_to_geopackage=False):
        if not path.exists():
            raise FileNotFoundError(f"File not found: {path}")
        self.db = ThreediDatabase(path)
        self._epsg_code = None  # for reproject()

        version = self.get_version()
        if version < MIN_SQLITE_VERSION:
            if upgrade:
                self.upgrade(convert_to_geopackage=convert_to_geopackage)
            else:
                raise SchematisationError(f"Too old sqlite version {version}.")

    def get_version(self) -> int:
        # check version
        schema = ModelSchema(self.db)
        return schema.get_version()

    def upgrade(self, convert_to_geopackage=False):
        schema = ModelSchema(self.db)
        schema.upgrade(
            backup=False, set_views=False, convert_to_geopackage=convert_to_geopackage
        )

    @contextmanager
    def get_session(self) -> ContextManager[Session]:
        """A context manager that yields an SQLAlchemy session.

        The session is closed af the context manager exit. No commit or rollback
        is done. It is meant for read-only access, but writing is not prohibited.

        Returns:
          SQLAlchemy.orm.Session
        """
        session = self.db.get_session(query_cls=NumpyQuery)
        try:
            yield session
        finally:
            session.close()

    def get_settings(self) -> dict:
        """Return the settings relevant for makegrid and maketables.

        Returns:
           dict with epsg_code, model_name, grid_settings, make_table_settings
        """

        with self.get_session() as session:
            model_settings = session.query(models.ModelSettings).order_by("id").first()
            if (
                model_settings.use_groundwater_flow
                or model_settings.use_groundwater_storage
            ):
                groundwater = _object_as_dict(session.query(models.GroundWater).one())
            else:
                groundwater = {}
            if model_settings.use_interflow:
                interflow = _object_as_dict(session.query(models.Interflow).one())
            else:
                interflow = {}
            if model_settings.use_simple_infiltration:
                infiltration = _object_as_dict(
                    session.query(models.SimpleInfiltration).one()
                )
            else:
                infiltration = {}
            if model_settings.use_vegetation_drag_2d:
                vegetation_drag = _object_as_dict(
                    session.query(models.VegetationDrag).one()
                )
            else:
                vegetation_drag = {}
            if model_settings.use_interception:
                interception = _object_as_dict(session.query(models.Interception).one())
            else:
                interception = {}
            model_settings = _object_as_dict(model_settings)

        # record if there is a DEM file to be expected
        # Note: use_2d_flow only determines whether there are flow lines
        model_settings["use_2d"] = bool(model_settings["dem_file"])

        # set/adapt initialization types to include information about file presence
        NO_AGG = InitializationType.NO_AGG
        AVERAGE = InitializationType.AVERAGE
        _set_initialization_type(
            model_settings,
            "friction_coefficient",
            default=AVERAGE if model_settings["friction_averaging"] else NO_AGG,
        )
        _set_initialization_type(
            interception,
            "interception",
            file_field="interception_file",
            type_field="interception_type",
            default=NO_AGG,
        )
        if interflow:
            _set_initialization_type(interflow, "porosity", default=NO_AGG)
            _set_initialization_type(
                interflow, "hydraulic_conductivity", default=AVERAGE
            )
        if infiltration:
            _set_initialization_type(infiltration, "infiltration_rate", default=NO_AGG)
            _set_initialization_type(
                infiltration, "max_infiltration_volume", default=NO_AGG
            )
        if groundwater:
            # default is what the user supplied (MIN/MAX/AVERAGE)

            _set_initialization_type(
                groundwater,
                "groundwater_impervious_layer_level",
                type_field="groundwater_impervious_layer_level_aggregation",
            )
            _set_initialization_type(
                groundwater,
                "phreatic_storage_capacity",
                type_field="phreatic_storage_capacity_aggregation",
            )
            _set_initialization_type(groundwater, "equilibrium_infiltration_rate")
            _set_initialization_type(
                groundwater,
                "initial_infiltration_rate",
                type_field="initial_infiltration_rate_aggregation",
            )
            _set_initialization_type(
                groundwater,
                "infiltration_decay_period",
                type_field="infiltration_decay_period_aggregation",
            )
            _set_initialization_type(
                groundwater,
                "groundwater_hydraulic_conductivity",
                type_field="groundwater_hydraulic_conductivity_aggregation",
            )

        if vegetation_drag:
            _set_initialization_type(
                vegetation_drag, "vegetation_height", default=NO_AGG
            )
            _set_initialization_type(
                vegetation_drag, "vegetation_stem_count", default=NO_AGG
            )
            _set_initialization_type(
                vegetation_drag, "vegetation_stem_diameter", default=NO_AGG
            )
            _set_initialization_type(
                vegetation_drag, "vegetation_drag_coefficient", default=NO_AGG
            )
        # Copy Simulation Template Settings to model_settings dict
        template_settings = _object_as_dict(
            session.query(models.SimulationTemplateSettings).order_by("id").first()
        )
        model_settings["name"] = template_settings["name"]
        model_settings["use_0d_inflow"] = template_settings["use_0d_inflow"]

        grid_settings = GridSettings.from_dict(model_settings)
        tables_settings = TablesSettings.from_dict(
            {
                **groundwater,
                **interflow,
                **infiltration,
                **vegetation_drag,
                **model_settings,
                **interception,
            }
        )
        return {
            "epsg_code": model_settings["epsg_code"],
            "model_name": model_settings["name"],
            "grid_settings": grid_settings,
            "tables_settings": tables_settings,
        }

    @property
    def epsg_code(self) -> int:
        if self._epsg_code is None:
            self._epsg_code = self.get_settings()["epsg_code"]
        return self._epsg_code

    @epsg_code.setter
    def epsg_code(self, value: int):
        self._epsg_code = value

    def reproject(self, geometries: np.ndarray) -> np.ndarray:
        """Reproject geometries from 4326 to the EPSG in the settings.

        Notes:
          shapely+pyproj is approx 2x faster than spatialite

        Args:
          geometries (ndarray of shapely.Geometry): geometries in EPSG 4326
        """
        target_epsg = self.epsg_code
        func = _get_reproject_func(SOURCE_EPSG, target_epsg)
        return shapely.transform(geometries, func)

    def get_surfaces(self) -> Surfaces:
        with self.get_session() as session:
            arr = (
                session.query(
                    models.Surface.id.label("surface_id"),
                    models.Surface.code,
                    models.Surface.display_name,
                    models.Surface.area,
                    models.Surface.geom,
                    models.SurfaceParameter.outflow_delay,
                    models.SurfaceParameter.surface_layer_thickness,
                    models.SurfaceParameter.infiltration,
                    models.SurfaceParameter.max_infiltration_capacity,
                    models.SurfaceParameter.min_infiltration_capacity,
                    models.SurfaceParameter.infiltration_decay_constant,
                    models.SurfaceParameter.infiltration_recovery_constant,
                    models.ConnectionNode.id.label("connection_node_id"),
                    models.ConnectionNode.the_geom.label("connection_node_the_geom"),
                    models.SurfaceMap.percentage,
                )
                .select_from(models.Surface)
                .join(models.SurfaceParameter)
                .join(
                    models.SurfaceMap, models.SurfaceMap.surface_id == models.Surface.id
                )
                .join(
                    models.ConnectionNode,
                    models.SurfaceMap.connection_node_id == models.ConnectionNode.id,
                )
                .order_by(models.Surface.id)
                .as_structarray()
            )

        # reproject
        arr["geom"] = self.reproject(arr["geom"])
        arr["connection_node_the_geom"] = self.reproject(
            arr["connection_node_the_geom"]
        )

        return Surfaces(
            id=np.arange(0, len(arr["surface_id"] + 1), dtype=int),
            **{name: arr[name] for name in arr.dtype.names},
        )

    def get_boundary_conditions_1d(self) -> BoundaryConditions1D:
        """Return BoundaryConditions1D"""
        with self.get_session() as session:
            arr = (
                session.query(
                    models.BoundaryCondition1D.id,
                    models.BoundaryCondition1D.boundary_type,
                    models.BoundaryCondition1D.connection_node_id,
                )
                .order_by(models.BoundaryCondition1D.id)
                .as_structarray()
            )

        # transform to a BoundaryConditions1D object
        return BoundaryConditions1D(**{name: arr[name] for name in arr.dtype.names})

    def get_boundary_conditions_2d(self) -> BoundaryConditions2D:
        """Return BoundaryConditions2D"""
        with self.get_session() as session:
            arr = (
                session.query(
                    models.BoundaryConditions2D.id,
                    models.BoundaryConditions2D.boundary_type,
                    models.BoundaryConditions2D.the_geom,
                )
                .order_by(models.BoundaryConditions2D.id)
                .as_structarray()
            )
        arr["the_geom"] = self.reproject(arr["the_geom"])

        # transform to a BoundaryConditions1D object
        return BoundaryConditions2D(**{name: arr[name] for name in arr.dtype.names})

    def get_channels(self) -> Channels:
        """Return Channels"""
        cols = [
            models.Channel.the_geom,
            models.Channel.dist_calc_points,
            models.Channel.id,
            models.Channel.code,
            models.Channel.connection_node_start_id,
            models.Channel.connection_node_end_id,
            models.Channel.calculation_type,
            models.Channel.display_name,
            models.Channel.zoom_category,
            models.Channel.exchange_thickness,
            models.Channel.hydraulic_conductivity_out,
            models.Channel.hydraulic_conductivity_in,
        ]

        with self.get_session() as session:
            arr = session.query(*cols).order_by(models.Channel.id).as_structarray()

        arr["the_geom"] = self.reproject(arr["the_geom"])
        # map "old" calculation types (100, 101, 102, 105) to (0, 1, 2, 5)
        arr["calculation_type"][arr["calculation_type"] >= 100] -= 100
        arr["hydraulic_conductivity_out"] /= DAY_IN_SECONDS
        arr["hydraulic_conductivity_in"] /= DAY_IN_SECONDS

        # transform to a Channels object
        return Channels(**{name: arr[name] for name in arr.dtype.names})

    def get_connection_nodes(self) -> ConnectionNodes:
        """Return ConnectionNodes (which are enriched using the manhole table)"""
        cols = [
            models.ConnectionNode.the_geom,
            models.ConnectionNode.id,
            models.ConnectionNode.code,
            models.ConnectionNode.storage_area,
            models.ConnectionNode.initial_waterlevel,
            models.Manhole.id.label("manhole_id"),
            models.Manhole.calculation_type,
            models.Manhole.bottom_level,
            models.Manhole.drain_level,
            models.Manhole.manhole_indicator,
            models.Manhole.surface_level,
            models.Manhole.shape,
            models.Manhole.width,
            models.Manhole.display_name,
            models.Manhole.zoom_category,
            models.Manhole.exchange_thickness,
            models.Manhole.hydraulic_conductivity_out,
            models.Manhole.hydraulic_conductivity_in,
        ]

        with self.get_session() as session:
            arr = (
                session.query(*cols)
                .join(models.ConnectionNode.manholes, isouter=True)
                .order_by(models.ConnectionNode.id)
                .as_structarray()
            )

        arr["the_geom"] = self.reproject(arr["the_geom"])

        # replace -9999.0 with NaN in initial_waterlevel
        arr["initial_waterlevel"][arr["initial_waterlevel"] == -9999.0] = np.nan
        arr["hydraulic_conductivity_out"] /= DAY_IN_SECONDS
        arr["hydraulic_conductivity_in"] /= DAY_IN_SECONDS

        return ConnectionNodes(**{name: arr[name] for name in arr.dtype.names})

    def get_cross_section_definitions(self) -> CrossSectionDefinitions:
        """Return CrossSectionDefinitions"""
        with self.get_session() as session:
            arr = (
                session.query(
                    models.CrossSectionDefinition.id,
                    models.CrossSectionDefinition.code,
                    models.CrossSectionDefinition.shape,
                    models.CrossSectionDefinition.width,
                    models.CrossSectionDefinition.height,
                    models.CrossSectionDefinition.friction_values,
                    models.CrossSectionDefinition.vegetation_stem_densities,
                    models.CrossSectionDefinition.vegetation_stem_diameters,
                    models.CrossSectionDefinition.vegetation_heights,
                    models.CrossSectionDefinition.vegetation_drag_coefficients,
                )
                .order_by(models.CrossSectionDefinition.id)
                .as_structarray()
            )

        # map shape 10 to 1 (circle) to match CrossSectionShape enum
        arr["shape"][arr["shape"] == 10] = 1
        # map shape 11 to 5 (tabulated rectangle) to match CrossSectionShape enum
        arr["shape"][arr["shape"] == 11] = 5
        # map shape 12 to 6 (tabulated trapezium) to match CrossSectionShape enum
        arr["shape"][arr["shape"] == 12] = 6

        # transform to a CrossSectionDefinitions object
        return CrossSectionDefinitions(**{name: arr[name] for name in arr.dtype.names})

    def get_cross_section_locations(self) -> CrossSectionLocations:
        """Return CrossSectionLocations"""
        with self.get_session() as session:
            arr = (
                session.query(
                    models.CrossSectionLocation.id,
                    models.CrossSectionLocation.code,
                    models.CrossSectionLocation.the_geom,
                    models.CrossSectionLocation.definition_id,
                    models.CrossSectionLocation.channel_id,
                    models.CrossSectionLocation.reference_level,
                    models.CrossSectionLocation.bank_level,
                    models.CrossSectionLocation.friction_type,
                    models.CrossSectionLocation.friction_value,
                    models.CrossSectionLocation.vegetation_stem_density,
                    models.CrossSectionLocation.vegetation_stem_diameter,
                    models.CrossSectionLocation.vegetation_height,
                    models.CrossSectionLocation.vegetation_drag_coefficient,
                )
                .order_by(models.CrossSectionLocation.id)
                .as_structarray()
            )

        arr["the_geom"] = self.reproject(arr["the_geom"])

        # transform to a CrossSectionLocations object
        return CrossSectionLocations(**{name: arr[name] for name in arr.dtype.names})

    def get_culverts(self) -> Culverts:
        """Return Culverts"""
        with self.get_session() as session:
            arr = (
                session.query(
                    models.Culvert.id,
                    models.Culvert.code,
                    models.Culvert.the_geom,
                    models.Culvert.dist_calc_points,
                    models.Culvert.connection_node_start_id,
                    models.Culvert.connection_node_end_id,
                    models.Culvert.calculation_type,
                    models.Culvert.cross_section_definition_id,
                    models.Culvert.invert_level_start_point,
                    models.Culvert.invert_level_end_point,
                    models.Culvert.discharge_coefficient_negative,
                    models.Culvert.discharge_coefficient_positive,
                    models.Culvert.friction_type,
                    models.Culvert.friction_value,
                    models.Culvert.display_name,
                    models.Culvert.zoom_category,
                )
                .order_by(models.Culvert.id)
                .as_structarray()
            )

        arr["the_geom"] = self.reproject(arr["the_geom"])

        # map friction_type 4 to friction_type 2 to match crosssectionlocation enum
        arr["friction_type"][arr["friction_type"] == 4] = 2

        # When no calculation type is provides we default to isolated
        arr["calculation_type"][
            arr["calculation_type"] == -9999
        ] = LineType.LINE_1D_ISOLATED
        # map "old" calculation types (100, 101, 102, 105) to (0, 1, 2, 5)
        arr["calculation_type"][arr["calculation_type"] >= 100] -= 100

        # transform to a CrossSectionLocations object
        return Culverts(**{name: arr[name] for name in arr.dtype.names})

    def get_exchange_lines(self) -> ExchangeLines:
        with self.get_session() as session:
            arr = (
                session.query(
                    models.ExchangeLine.id,
                    models.ExchangeLine.channel_id,
                    models.ExchangeLine.the_geom,
                    models.ExchangeLine.exchange_level,
                )
                .order_by(models.ExchangeLine.id)
                .as_structarray()
            )

        arr["the_geom"] = self.reproject(arr["the_geom"])

        # transform to a Channels object
        return ExchangeLines(**{name: arr[name] for name in arr.dtype.names})

    def get_grid_refinements(self) -> GridRefinements:
        """Return Gridrefinement and GridRefinementArea concatenated into one array."""
        with self.get_session() as session:
            arr1 = (
                session.query(
                    models.GridRefinement.the_geom,
                    models.GridRefinement.id,
                    models.GridRefinement.code,
                    models.GridRefinement.display_name,
                    models.GridRefinement.refinement_level,
                )
                .filter(
                    models.GridRefinement.the_geom.isnot(None),
                    models.GridRefinement.refinement_level.isnot(None),
                )
                .order_by(models.GridRefinement.id)
                .as_structarray()
            )
            arr2 = (
                session.query(
                    models.GridRefinementArea.the_geom,
                    models.GridRefinementArea.id,
                    models.GridRefinementArea.code,
                    models.GridRefinementArea.display_name,
                    models.GridRefinementArea.refinement_level,
                )
                .filter(
                    models.GridRefinementArea.the_geom.isnot(None),
                    models.GridRefinementArea.refinement_level.isnot(None),
                )
                .order_by(models.GridRefinementArea.id)
                .as_structarray()
            )
            arr = np.concatenate((arr1, arr2))

        # reproject
        arr["the_geom"] = self.reproject(arr["the_geom"])
        arr["id"] = np.arange(len(arr["refinement_level"]))

        return GridRefinements(**{name: arr[name] for name in arr.dtype.names})

    def get_dem_average_areas(self) -> DemAverageAreas:
        """Return DemAverageAreas"""
        with self.get_session() as session:
            arr = (
                session.query(
                    models.DemAverageArea.id,
                    models.DemAverageArea.the_geom,
                )
                .order_by(models.DemAverageArea.id)
                .as_structarray()
            )
            arr["the_geom"] = self.reproject(arr["the_geom"])

        return DemAverageAreas(**{name: arr[name] for name in arr.dtype.names})

    def get_obstacles(self) -> Obstacles:
        with self.get_session() as session:
            arr = (
                session.query(
                    models.Obstacle.the_geom,
                    models.Obstacle.id,
                    models.Obstacle.crest_level,
                )
                .order_by(models.Obstacle.id)
                .as_structarray()
            )

        # reproject
        arr["the_geom"] = self.reproject(arr["the_geom"])

        return Obstacles(**{name: arr[name] for name in arr.dtype.names})

    def get_orifices(self) -> Orifices:
        """Return Orifices"""
        with self.get_session() as session:
            arr = (
                session.query(
                    models.Orifice.id,
                    models.Orifice.code,
                    models.Orifice.connection_node_start_id,
                    models.Orifice.connection_node_end_id,
                    models.Orifice.crest_level,
                    models.Orifice.crest_type,
                    models.Orifice.cross_section_definition_id,
                    models.Orifice.discharge_coefficient_negative,
                    models.Orifice.discharge_coefficient_positive,
                    models.Orifice.friction_type,
                    models.Orifice.friction_value,
                    models.Orifice.display_name,
                    models.Orifice.zoom_category,
                    models.Orifice.sewerage,
                )
                .order_by(models.Orifice.id)
                .as_structarray()
            )

        # map friction_type 4 to friction_type 2 to match crosssectionlocation enum
        arr["friction_type"][arr["friction_type"] == 4] = 2

        return Orifices(**{name: arr[name] for name in arr.dtype.names})

    def get_pipes(self) -> Pipes:
        """Return Pipes"""
        cols = [
            models.Pipe.id,
            models.Pipe.code,
            models.Pipe.sewerage_type,
            models.Pipe.calculation_type,
            models.Pipe.invert_level_start_point,
            models.Pipe.invert_level_end_point,
            models.Pipe.friction_type,
            models.Pipe.friction_value,
            models.Pipe.dist_calc_points,
            models.Pipe.connection_node_start_id,
            models.Pipe.connection_node_end_id,
            models.Pipe.cross_section_definition_id,
            models.Pipe.display_name,
            models.Pipe.zoom_category,
            models.Pipe.material,
            models.Pipe.exchange_thickness,
            models.Pipe.hydraulic_conductivity_out,
            models.Pipe.hydraulic_conductivity_in,
        ]

        with self.get_session() as session:
            arr = session.query(*cols).order_by(models.Pipe.id).as_structarray()

        # map friction_type 4 to friction_type 2 to match crosssectionlocation enum
        arr["friction_type"][arr["friction_type"] == 4] = 2
        arr["hydraulic_conductivity_out"] /= DAY_IN_SECONDS
        arr["hydraulic_conductivity_in"] /= DAY_IN_SECONDS

        # transform to a Pipes object
        return Pipes(**{name: arr[name] for name in arr.dtype.names})

    def get_pumps(self) -> Pumps:
        with self.get_session() as session:
            arr = (
                session.query(
                    models.Pumpstation.id,
                    models.Pumpstation.code,
                    models.Pumpstation.capacity,
                    models.Pumpstation.connection_node_start_id,
                    models.Pumpstation.connection_node_end_id,
                    models.Pumpstation.type_,
                    models.Pumpstation.start_level,
                    models.Pumpstation.lower_stop_level,
                    models.Pumpstation.upper_stop_level,
                    models.Pumpstation.display_name,
                    models.Pumpstation.zoom_category,
                )
                .order_by(models.Pumpstation.id)
                .as_structarray()
            )

        # Pump capicity is entered as L/s but we need m3/s.
        arr["capacity"] = arr["capacity"] / 1000

        # transform to a Pumps object
        return Pumps(**{name: arr[name] for name in arr.dtype.names})

    def get_weirs(self) -> Weirs:
        """Return Weirs"""
        with self.get_session() as session:
            arr = (
                session.query(
                    models.Weir.id,
                    models.Weir.code,
                    models.Weir.connection_node_start_id,
                    models.Weir.connection_node_end_id,
                    models.Weir.crest_level,
                    models.Weir.crest_type,
                    models.Weir.cross_section_definition_id,
                    models.Weir.discharge_coefficient_negative,
                    models.Weir.discharge_coefficient_positive,
                    models.Weir.friction_type,
                    models.Weir.friction_value,
                    models.Weir.display_name,
                    models.Weir.zoom_category,
                    models.Weir.sewerage,
                )
                .order_by(models.Weir.id)
                .as_structarray()
            )

        # map friction_type 4 to friction_type 2 to match crosssectionlocation enum
        arr["friction_type"][arr["friction_type"] == 4] = 2

        return Weirs(**{name: arr[name] for name in arr.dtype.names})

    def get_windshieldings(self) -> Windshieldings:
        with self.get_session() as session:
            arr = (
                session.query(
                    models.Windshielding.id,
                    models.Windshielding.channel_id,
                    models.Windshielding.north,
                    models.Windshielding.northeast,
                    models.Windshielding.east,
                    models.Windshielding.southeast,
                    models.Windshielding.south,
                    models.Windshielding.southwest,
                    models.Windshielding.west,
                    models.Windshielding.northwest,
                )
                .order_by(models.Windshielding.id)
                .as_structarray()
            )

        return Windshieldings(**{name: arr[name] for name in arr.dtype.names})

    def get_potential_breaches(self) -> PotentialBreaches:
        # potential breaches are available from schema version 211
        if self.get_version() < 211:
            return PotentialBreaches(id=[])

        cols = [
            models.PotentialBreach.id,
            models.PotentialBreach.code,
            models.PotentialBreach.display_name,
            models.PotentialBreach.the_geom,
            models.PotentialBreach.channel_id,
        ]

        if self.get_version() >= 212:
            cols += [
                models.PotentialBreach.exchange_level,
                models.PotentialBreach.maximum_breach_depth,
                models.PotentialBreach.levee_material,
            ]

        with self.get_session() as session:
            arr = (
                session.query(*cols)
                .order_by(models.PotentialBreach.id)
                .as_structarray()
            )

        # reproject
        arr["the_geom"] = self.reproject(arr["the_geom"])

        return PotentialBreaches(**{name: arr[name] for name in arr.dtype.names})


# Constructing a Transformer takes quite long, so we use caching here. The
# function is deterministic so this doesn't have any side effects.
@lru_cache(maxsize=8)
def _get_reproject_func(source_epsg: int, target_epsg: int) -> Callable:
    transformer = Transformer.from_crs(
        CRS.from_epsg(source_epsg), CRS.from_epsg(target_epsg), always_xy=True
    )

    def func(coords):
        if coords.shape[0] == 0:
            return coords
        x, y = transformer.transform(coords[:, 0], coords[:, 1])
        return np.array([x, y]).T

    return func



import numpy as np
import pytest
from numpy.testing import assert_array_equal

from threedigrid_builder.base import Nodes, Pumps
from threedigrid_builder.constants import ContentType
from threedigrid_builder.exceptions import SchematisationError


@pytest.fixture
def nodes():
    CN = ContentType.TYPE_V2_CONNECTION_NODES
    CH = ContentType.TYPE_V2_CHANNEL
    return Nodes(
        id=[0, 1, 3, 4],
        content_type=[CH, CN, CN, CH],
        content_pk=[21, 21, 33, 42],
        coordinates=[(0, 21), (1, 25), (2, 33), (3, 42)],
        dmax=[1.0, 2.0, 3.0, 4.0],
    )


@pytest.fixture
def pumps():
    return Pumps(
        id=[2, 35],
        line=[(1, 3), (1, -9999)],
        connection_node_start_id=[21, 21],
        connection_node_end_id=[33, -9999],
    )


def test_renumber(pumps):
    pumps.renumber()
    assert_array_equal(pumps.id, [0, 1])
    assert_array_equal(pumps.content_pk, [2, 35])


def test_set_lines(pumps, nodes):
    pumps.line[:] = -9999
    pumps.set_lines(nodes)
    assert_array_equal(pumps.line, [(1, 3), (1, -9999)])


@pytest.mark.parametrize(
    "start,end",
    [
        ([-9999, 21], [33, 33]),
        ([42, 21], [33, 33]),
        ([21, 21], [33, 42]),
    ],
)
def test_set_lines_error(start, end, pumps, nodes):
    pumps.connection_node_start_id[:] = start
    pumps.connection_node_end_id[:] = end
    with pytest.raises(SchematisationError):
        pumps.set_lines(nodes)


def test_set_node_data(pumps, nodes):
    pumps.set_node_data(nodes)
    assert_array_equal(pumps.line_coords, [(1, 25, 2, 33), (1, 25, np.nan, np.nan)])
    assert_array_equal(pumps.coordinates, [(1.5, 29.0), (1.0, 25.0)])
    assert_array_equal(pumps.bottom_level, [2.0, 2.0])



import itertools
from collections import defaultdict, deque

import numpy as np
import shapely

from threedigrid_builder.base import Nodes
from threedigrid_builder.constants import CalculationType, ContentType
from threedigrid_builder.exceptions import SchematisationError

from .linear import counts_to_ranges

# This is for judging from what cell to what cell a channel-cell edge intersection goes:
VPOINT_EPSILON = 1e-7


__all__ = ["embed_nodes", "embed_linear_objects"]


def take(arr, idx):
    """Take values at idx from arr, filling NaN where idx is -9999"""
    result = np.take(arr, idx, mode="clip")
    result[idx == -9999] = np.nan
    return result


def embed_nodes(grid, embedded_node_id_counter):
    """Integrate embedded connection nodes into the 2D cells.

    This changes the grid (nodes as well as lines) inplace.

    Returned are the removed (embedded) nodes.
    """
    is_embedded = (grid.nodes.calculation_type == CalculationType.EMBEDDED) & (
        grid.nodes.content_type == ContentType.TYPE_V2_CONNECTION_NODES
    )
    embedded_idx = np.where(is_embedded)[0]
    n_embedded = len(embedded_idx)
    if n_embedded == 0:
        return Nodes(id=[])

    # The query returns 2 1D arrays: one with indices into the embedded nodes
    # and one with indices into the cells.
    idx = grid.cell_tree.query(shapely.points(grid.nodes.coordinates[embedded_idx]))
    # Address cases of no or multiple cells per embedded node
    unique_embedded_idx, first_unique_index = np.unique(idx[0], return_index=True)
    # multiple cells: just take the first one
    idx = idx[:, first_unique_index]
    combs_embedded_idx = embedded_idx[idx[0]]
    combs_cell_idx = idx[1]
    if len(unique_embedded_idx) != n_embedded:
        missing_idx = embedded_idx[~np.isin(embedded_idx, combs_embedded_idx)]
        missing_cn_id = grid.nodes.content_pk[missing_idx].tolist()
        raise SchematisationError(
            f"Embedded connection nodes {missing_cn_id} are outside the 2D cells"
        )

    # create a mapping with new node ids on the position of the old node indices
    new_ids = np.full_like(grid.nodes.id, fill_value=-9999)
    new_ids[~is_embedded] = np.arange(len(grid.nodes) - n_embedded)
    new_ids[combs_embedded_idx] = grid.nodes.id[combs_cell_idx]

    # cut out the embedded nodes
    embedded_nodes = grid.nodes[combs_embedded_idx]
    embedded_nodes.id[:] = list(
        itertools.islice(embedded_node_id_counter, len(embedded_nodes))
    )
    embedded_nodes.node_type[:] = -9999  # clear the node_type; these are not calc nodes
    embedded_nodes.embedded_in[:] = new_ids[combs_cell_idx]
    grid.nodes = grid.nodes[~is_embedded]

    # map node indices (fixes node replacement and node renumbering in one go)
    grid.nodes.id[:] = np.arange(len(grid.nodes))
    grid.lines.line[:] = np.take(new_ids, grid.lines.line)
    if grid.pumps is not None:
        mask = grid.pumps.line != -9999
        grid.pumps.line[mask] = np.take(new_ids, grid.pumps.line[mask])
    if grid.surface_maps is not None:
        grid.surface_maps.cci[:] = np.take(new_ids, grid.surface_maps.cci)

    # check if there are no cells connecting to itself
    is_self_connected = grid.lines.line[:, 0] == grid.lines.line[:, 1]
    if np.any(is_self_connected):
        mask = np.isin(
            embedded_nodes.embedded_in, grid.lines.line[is_self_connected, 0]
        )
        raise SchematisationError(
            f"Embedded connection nodes {embedded_nodes.content_pk[mask].tolist()} connect to "
            f"one another within the same 2D cell."
        )

    return embedded_nodes


class EmbeddedObjects:
    """Embedded linear objects

    Attributes:
        objects (LinearObjects): the linear objects (input)
        vpoint_s (ndarray of float): positions of velocity points along objects
        vpoint_ch_idx (ndarray of int): index into objects corresponding to vpoint_s
        vpoint_line (length-2 iterable of ndarrays of int): cell ids belonging
          to the velocity points. Each velocity point has two cell ids because the
          point is on a cell edge by definition.
    """

    def __init__(
        self, objects, vpoint_s, vpoint_ch_idx, vpoint_line=None, vpoint_line_s=None
    ):
        self.objects = objects
        self.vpoint_s = vpoint_s
        self.vpoint_ch_idx = vpoint_ch_idx
        self.vpoint_line = vpoint_line
        self.vpoint_line_s = vpoint_line_s

    @property
    def counts(self):
        """The number of lines per channel"""
        return np.bincount(self.vpoint_ch_idx, minlength=len(self.objects))

    def __repr__(self):
        return (
            f"<{self.__class__.__name__} object with "
            f"{len(self.objects)} {self.objects.__class__.__name__} and "
            f"{len(self.vpoint_s)} velocity points>"
        )

    def __getitem__(self, index):
        """Get velocity points at ``index``"""
        return self.__class__(
            self.objects,
            self.vpoint_s[index],
            self.vpoint_ch_idx[index],
            self.vpoint_line[:, index] if self.vpoint_line is not None else None,
            self.vpoint_line_s[index] if self.vpoint_line_s is not None else None,
        )

    def delete(self, index):
        """Delete velocity points at ``index``"""
        if self.vpoint_line is not None:
            vpoint_line = np.delete(self.vpoint_line, index, axis=1)
        else:
            vpoint_line = None
        if self.vpoint_line_s is not None:
            vpoint_line_s = np.delete(self.vpoint_line_s, index)
        else:
            vpoint_line_s = None
        return self.__class__(
            self.objects,
            np.delete(self.vpoint_s, index),
            np.delete(self.vpoint_ch_idx, index),
            vpoint_line,
            vpoint_line_s,
        )

    def insert(
        self, where, vpoint_s, vpoint_ch_idx, vpoint_line=-9999, vpoint_line_s=np.nan
    ):
        """Insert velocity points at ``where``"""
        if self.vpoint_line is not None:
            vpoint_line = np.insert(self.vpoint_line, where, vpoint_line, axis=1)
        else:
            vpoint_line = None
        if self.vpoint_line_s is not None:
            vpoint_line_s = np.insert(self.vpoint_line_s, where, vpoint_line_s, axis=1)
        else:
            vpoint_line_s = None
        return self.__class__(
            self.objects,
            np.insert(self.vpoint_s, where, vpoint_s),
            np.insert(self.vpoint_ch_idx, where, vpoint_ch_idx),
            vpoint_line,
            vpoint_line_s,
        )

    @classmethod
    def from_cell_tree(cls, objects, cell_tree):
        """Construct the velocity points in the embedded objects from a cell tree."""
        # The query returns 2 1D arrays: one with indices into the objects
        # and one with indices into the cells.
        idx = cell_tree.query(objects.the_geom, "intersects")
        # Get the intersections between objects and cell edges: these are the vpoints
        _points = shapely.intersection(
            objects.the_geom[idx[0]],
            shapely.get_exterior_ring(cell_tree.geometries[idx[1]]),
        )
        # Get unique points per channel (there will be duplicates).
        # This also converts linestring intersections; every vertex becomes a point.
        multipoints = shapely.extract_unique_points(
            shapely.geometrycollections(_points, indices=idx[0])
        )
        # Flatten into a 'ragged array' structure (points + indices into objects)
        vpoints, vpoint_ch_idx = shapely.get_parts(multipoints, return_index=True)
        # Measure the location along the objects
        vpoint_s = shapely.line_locate_point(objects.the_geom[vpoint_ch_idx], vpoints)
        return cls(objects, vpoint_s, vpoint_ch_idx)

    def get_nodes(self, node_id_counter):
        """Get the (virtual) Nodes corresponding to the velocity points"""
        ch_start, ch_end = counts_to_ranges(self.counts)
        node_s = (
            np.delete(self.vpoint_s, ch_start) + np.delete(self.vpoint_s, ch_end - 1)
        ) / 2
        self.vpoint_line_s = self.vpoint_s.copy()
        int_nodes = np.delete(np.arange(len(self.vpoint_line_s)), ch_start)
        self.vpoint_line_s[int_nodes] -= node_s
        node_ch_idx = np.delete(self.vpoint_ch_idx, ch_start)
        node_point = shapely.line_interpolate_point(
            self.objects.the_geom[node_ch_idx], node_s
        )
        node_cell_id = np.delete(self.vpoint_line[0], ch_start)
        return Nodes(
            id=itertools.islice(node_id_counter, len(node_ch_idx)),
            coordinates=shapely.get_coordinates(node_point),
            content_type=self.objects.content_type,
            content_pk=self.objects.id[node_ch_idx],
            calculation_type=CalculationType.EMBEDDED,
            s1d=node_s,
            embedded_in=node_cell_id,
        )

    def sort(self):
        """Sort the velocity points by channel and then by position on the channel"""
        return self[np.lexsort((self.vpoint_s, self.vpoint_ch_idx))]

    def clean_vpoints_channel_ending(self):
        """Solve an edge case where a channel ending is at a cell edge.

        The EmbeddedObjects construction erroneously puts velocity points at channel
        endings if the ending is precisely at the cell edge. These are filtered out.
        """
        ch_lengths = shapely.length(self.objects.the_geom)
        return self[
            (self.vpoint_s > 0.0) & (self.vpoint_s < ch_lengths[self.vpoint_ch_idx])
        ]

    def set_cell_ids(self, cell_tree):
        """Compute 2 cell ids (from and to) for each velocity point (line), inplace.

        This also deals with edge cases in which a line lies tangent to a cell edge. If
        a line goes exactly over a cell edge, this cell is not included. In other words,
        an embedded channel is only embedded in a cell if the channel intersects with
        the cell interior.
        """
        # For each vpoint, see from which to which cell it goes. This is done by
        # going a bit before and after the vpoint and using the cell_tree.
        pnt_a = shapely.line_interpolate_point(
            self.objects.the_geom[self.vpoint_ch_idx], self.vpoint_s - VPOINT_EPSILON
        )
        pnt_b = shapely.line_interpolate_point(
            self.objects.the_geom[self.vpoint_ch_idx], self.vpoint_s + VPOINT_EPSILON
        )
        vp_idx_a, cell_idx_a = cell_tree.query(pnt_a)
        vp_idx_b, cell_idx_b = cell_tree.query(pnt_b)

        # List velocity points that are outside of the model.
        n_vpoints = len(self.vpoint_s)
        no_intersct_a = np.bincount(vp_idx_a, minlength=n_vpoints) == 0
        no_intersct_b = np.bincount(vp_idx_b, minlength=n_vpoints) == 0
        if np.any(no_intersct_a | no_intersct_b):
            bad_objects_ids = self.vpoint_ch_idx[no_intersct_a | no_intersct_b].tolist()
            raise SchematisationError(
                f"{self.objects.__class__.__name__} {sorted(set(bad_objects_ids))} are "
                f"not completely inside the 2D cell."
            )

        cell_idx_a, cell_idx_b = self._fix_tangent_lines(
            vp_idx_a, cell_idx_a, vp_idx_b, cell_idx_b
        )
        self.vpoint_line = np.array([cell_idx_a, cell_idx_b], dtype=int)

    def clean_vpoints_not_crossing(self):
        """Remove velocity points that only touch (but not cross) a cell edge"""
        line_touches = self.vpoint_line[0] == self.vpoint_line[1]
        if np.any(line_touches):
            return self[~line_touches]
        else:
            return self

    def merge_close_points(self, threshold):
        """Merge velocity points that are closer than ``threshold`` together"""
        line_ds = np.diff(self.vpoint_s)
        line_too_short = np.where(
            (line_ds < threshold) & (np.diff(self.vpoint_ch_idx) == 0)
        )[0]
        if len(line_too_short) == 0:
            return self
        # Get consecutive ranges of too short segments
        line_too_short += 1
        _is_start = (line_too_short - np.roll(line_too_short, 1)) > 1
        _is_start[0] = True
        _first = line_too_short[_is_start] - 1
        _last = line_too_short[np.roll(_is_start, -1)]
        # Adjust the first of each consecutive range (change vpoint_s and vpoint_line)
        copy = self[:]
        copy.vpoint_s[_first] = (copy.vpoint_s[_first] + copy.vpoint_s[_last]) / 2
        copy.vpoint_line[1, _first] = copy.vpoint_line[1, _last]
        # Delete the rest
        return copy.delete(line_too_short)

    def fix_zero_points(self):
        """Add a velocity point for objects without any"""
        ch_no_lines = np.where(self.counts == 0)[0]
        if len(ch_no_lines) == 0:
            return self

        insert_where = np.searchsorted(self.vpoint_ch_idx, ch_no_lines)
        return self.insert(
            insert_where,
            vpoint_s=0.5 * shapely.length(self.objects.the_geom[ch_no_lines]),
            vpoint_ch_idx=ch_no_lines,
        )

    def _fix_tangent_lines(self, vpoint_idx_a, cell_idx_a, vpoint_idx_b, cell_idx_b):
        """Fix velocity points whose line around it is on a cell edge, inplace.

        Args:
            vpoint_idx_a (ndarray of int): velocity point index
            cell_idx_a (ndarray of int): index of the cell before the velocity point
            vpoint_idx_b (ndarray of int): velocity point index
            cell_idx_b (ndarray of int): index of the cell after the velocity point

        Returns:
            tuple of cell_idx_a, cell_idx_b, each with same length as self.vpoint_s

        Example:
            There is a channel that goes from cell 6 to the edge between 6 and 8 and
            then to cell 8. This channel will give 2 velocity points:
            Vpoint 0 goes from cell 6 to 6/8 and vpoint 1 goes from cell 6/8 to 8

            >>> vpoint_s = [5., 11.]
            >>> vpoint_idx_a, vpoint_line_a = [0, 1, 1], [6, 6, 8]
            >>> vpoint_idx_b, vpoint_line_b = [0, 0, 1], [6, 8, 8]

            These vpoints will be merged into:

            >>> vpoint_s = [8.]  # halfway
            >>> vpoint_idx_a, cell_idx_a = [0], [6]
            >>> vpoint_idx_b, cell_idx_b = [0], [8]

        Notes:
            To solve this connectivity problem, we recast the data structure into a
            directed graph containing both the cells and the velocity points. A standard
            shortest path algorithm is then used to get the path that crosses the
            least amount of cells.
        """
        n_vpoints = len(self.vpoint_s)
        if len(cell_idx_a) == n_vpoints and len(cell_idx_b) == n_vpoints:
            return cell_idx_a, cell_idx_b

        # Check which vpoints could have multiple cell transitions
        vpoint_has_mult_a = np.bincount(vpoint_idx_a, minlength=n_vpoints) > 1
        vpoint_has_mult_b = np.bincount(vpoint_idx_b, minlength=n_vpoints) > 1
        mult_vpoint_idx = np.where(vpoint_has_mult_a | vpoint_has_mult_b)[0]
        mult_ch_idx = self.vpoint_ch_idx[mult_vpoint_idx]

        # List consecutive ranges of vpoints with multiple cell transitions
        is_first = ((mult_vpoint_idx - np.roll(mult_vpoint_idx, 1)) > 1) | (
            mult_ch_idx != np.roll(mult_ch_idx, 1)
        )
        is_first[0] = True
        is_last = np.roll(is_first, -1)

        # Iterate over consecutive ranges (and keep a global mask of vpoints to keep)
        vpoints_to_keep = np.ones(n_vpoints, dtype=bool)
        for _first, _last in zip(mult_vpoint_idx[is_first], mult_vpoint_idx[is_last]):
            vpoints_to_keep[_first : _last + 1] = False

            # Construct the graph (e.g. {"c2": ["v1", "v2"], "v2": ["c1"], ...}
            graph = defaultdict(list)
            mask_a = np.where((vpoint_idx_a >= _first) & (vpoint_idx_a <= _last))[0]
            for vpoint_idx, cell_idx in zip(vpoint_idx_a[mask_a], cell_idx_a[mask_a]):
                graph[f"c{cell_idx}"].append(f"v{vpoint_idx}")
            mask_b = np.where((vpoint_idx_b >= _first) & (vpoint_idx_b <= _last))[0]
            for vpoint_idx, cell_idx in zip(vpoint_idx_b[mask_b], cell_idx_b[mask_b]):
                graph[f"v{vpoint_idx}"].append(f"c{cell_idx}")

            # Get the shortest path (e.g. ['c2', 'v2', 'c1'])
            # Note: There could be multiple cell options to start/end with, loop over
            # combinations and take the shortest outcome.
            path = None
            for a, b in itertools.product(
                cell_idx_a[vpoint_idx_a == _first], cell_idx_b[vpoint_idx_b == _last]
            ):
                _path = shortest_path(graph, f"c{a}", f"c{b}")
                if path is None or len(_path) < len(path):
                    path = _path
            n_links = int(len(path) / 2)  # len(path) is always odd, but int() floors

            # Reduce the size of cell_idx_a/b, vpoint_idx_a/b
            vpoint_idx_a = np.delete(vpoint_idx_a, mask_a[n_links:])
            cell_idx_a = np.delete(cell_idx_a, mask_a[n_links:])
            vpoint_idx_b = np.delete(vpoint_idx_b, mask_b[n_links:])
            cell_idx_b = np.delete(cell_idx_b, mask_b[n_links:])

            # Overwrite the cell_idx_a/b, vpoint_idx_a/b to connect the right cells
            for i in range(n_links):
                vpoint = int(path[2 * i + 1][1:])
                vpoints_to_keep[vpoint] = True
                vpoint_idx_a[mask_a[i]] = vpoint
                cell_idx_a[mask_a[i]] = int(path[2 * i][1:])
                vpoint_idx_b[mask_b[i]] = vpoint
                cell_idx_b[mask_b[i]] = int(path[2 * i + 2][1:])

        self.vpoint_s = self.vpoint_s[vpoints_to_keep]
        self.vpoint_ch_idx = self.vpoint_ch_idx[vpoints_to_keep]
        return cell_idx_a, cell_idx_b


def embed_linear_objects(
    objects,
    cell_tree,
    embedded_cutoff_threshold,
    embedded_node_id_counter,
):
    """Create embedded nodes for linear objects

    All linear objects are expected to be of EMBEDDED calculation type.
    """
    if len(objects) == 0:
        return Nodes(id=[]), np.empty((0,)), np.empty((0,))

    if cell_tree is None or len(cell_tree) == 0:
        raise SchematisationError(
            f"{objects.__class__.__name__} {objects.id.tolist()} have an embedded "
            f"calculation type while there is no 2D domain."
        )

    embedded_objects = EmbeddedObjects.from_cell_tree(objects, cell_tree)
    embedded_objects = embedded_objects.clean_vpoints_channel_ending()
    embedded_objects = embedded_objects.sort()
    embedded_objects.set_cell_ids(cell_tree)
    embedded_objects = embedded_objects.clean_vpoints_not_crossing()
    if embedded_cutoff_threshold:
        embedded_objects = embedded_objects.merge_close_points(
            embedded_cutoff_threshold
        )
    embedded_objects = embedded_objects.fix_zero_points()

    # Now we create the virtual nodes
    embedded_nodes = embedded_objects.get_nodes(embedded_node_id_counter)

    return embedded_nodes, embedded_objects.vpoint_s, embedded_objects.vpoint_line_s


def _bfs(graph, start):
    """See shortest_path"""
    queue, enqueued = deque([(None, start)]), set([start])
    while queue:
        parent, n = queue.popleft()
        yield parent, n
        new = set(graph[n]) - enqueued
        enqueued |= new
        queue.extend([(n, child) for child in new])


def shortest_path(graph, start, end):
    """Find the shortest path in a directed graph using Breath First Search (BFS)

    Source:
       https://code.activestate.com/recipes/576675/, August 2021, MIT Licensed
    """
    if start == end:
        return [start]
    parents = {}
    for parent, child in _bfs(graph, start):
        parents[child] = parent
        if child == end:
            revpath = [end]
            while True:
                parent = parents[child]
                revpath.append(parent)
                if parent == start:
                    break
                child = parent
            return list(reversed(revpath))



from typing import Tuple

import numpy as np

from threedigrid_builder.base import Array
from threedigrid_builder.constants import ContentType
from threedigrid_builder.exceptions import SchematisationError

__all__ = ["Pumps"]


class Pump:
    id: int
    code: str
    capacity: float
    connection_node_start_id: int
    connection_node_end_id: int  # optional!
    type_: int
    start_level: float
    lower_stop_level: float
    upper_stop_level: float
    zoom_category: int
    display_name: str
    # sewerage
    # Generated fields:
    content_pk: int
    line: Tuple[int, int]
    line_coords: Tuple[float, float, float, float]
    coordinates: Tuple[float, float]
    bottom_level: float


class Pumps(Array[Pump]):
    def renumber(self):
        """Renumber the ids of the pumps with a sequential ID.

        The original ID is stored in content_pk.
        """
        self.content_pk[:] = self.id
        self.id[:] = np.arange(len(self))

    def set_lines(self, nodes):
        """Set the node ids into self.line by mapping connection_node_start/end_id."""
        mask = nodes.content_type == ContentType.TYPE_V2_CONNECTION_NODES
        node_ids = nodes.id[mask]
        cn_ids = nodes.content_pk[mask]

        if (np.diff(cn_ids) < 0).any():
            # sort by connection node id so that we can lookup
            sorter = np.argsort(cn_ids)
            node_ids = node_ids[sorter]
            cn_ids = cn_ids[sorter]

        # check which connection node ids are specified
        has_start = self.connection_node_start_id >= 0
        has_end = self.connection_node_end_id >= 0
        if not has_start.all():
            raise SchematisationError(
                f"Pumps {self.content_pk[~has_start].tolist()} have no "
                f"connection_node_start_id"
            )

        # lookup the corresponding index into node_ids
        start_idx = np.searchsorted(cn_ids, self.connection_node_start_id)
        end_idx = np.searchsorted(cn_ids, self.connection_node_end_id[has_end])

        # check if the connection node ids indeed exist
        start_is_ok = (
            cn_ids.take(start_idx, mode="clip") == self.connection_node_start_id
        )
        end_is_ok = (
            cn_ids.take(end_idx, mode="clip") == self.connection_node_end_id[has_end]
        )
        if not (start_is_ok.all() and end_is_ok.all()):
            raise SchematisationError(
                f"Pumps {self.content_pk[~(start_is_ok & end_is_ok)].tolist()} refer "
                f"to nonexisting connection nodes."
            )
        self.line[:, 0] = node_ids[start_idx]
        self.line[has_end, 1] = node_ids[end_idx]

    def set_node_data(self, nodes):
        """Set bottom_level, line_coords, and coordinates from node ids in self.line."""
        has_end = self.line[:, 1] != -9999
        start_node_idx = nodes.id_to_index(self.line[:, 0])
        end_node_idx = nodes.id_to_index(self.line[has_end, 1])

        # copy over the bottom levels
        self.bottom_level[:] = nodes.dmax[start_node_idx]

        # set the line coordinates
        self.line_coords[:, :2] = nodes.coordinates[start_node_idx]
        self.line_coords[has_end, 2:] = nodes.coordinates[end_node_idx]

        # set the coordinates
        self.coordinates[:] = self.line_coords[:, :2]
        # take the centroid for pumps that have 2 coordinates
        self.coordinates[has_end] = (
            self.coordinates[has_end] + self.line_coords[has_end, 2:]
        ) / 2



import numpy as np
import pytest
import shapely
from numpy.testing import assert_almost_equal, assert_equal

from threedigrid_builder.base import Lines, Nodes
from threedigrid_builder.constants import LineType
from threedigrid_builder.grid import Obstacles
from threedigrid_builder.grid.grid import Grid


@pytest.fixture
def lines():
    lines = Lines(
        id=range(10),
        kcu=[98, 98, 98, 98, 98, 99, 99, 99, 99, 99],
        line_coords=[
            [12.0, 12.0, 16.0, 12.0],
            [12.0, 16.0, 15.0, 15.0],
            [12.0, 16.0, 15.0, 17.0],
            [15.0, 15.0, 17.0, 15.0],
            [15.0, 17.0, 17.0, 17.0],
            [12.0, 12.0, 12.0, 16.0],
            [16.0, 12.0, 15.0, 15.0],
            [16.0, 12.0, 17.0, 15.0],
            [15.0, 15.0, 15.0, 17.0],
            [17.0, 15.0, 17.0, 17.0],
        ],
    )
    lines.fix_line_geometries()
    return lines


@pytest.fixture
def grid(lines):
    return Grid(nodes=Nodes(id=np.array([1])), lines=lines)


@pytest.fixture
def obstacles():
    return Obstacles(
        id=[1, 2],
        crest_level=[-0.1, -0.3],
        the_geom=shapely.linestrings(
            [[[12.0, 11.0], [17.5, 17.0]], [[13.0, 11.0], [13.0, 17.0]]]
        ),
    )


@pytest.fixture
def obstacles_no_intersect():
    return Obstacles(
        id=[1],
        crest_level=[-0.1],
        the_geom=shapely.linestrings(
            [[[11, 11], [11, 18], [18, 18], [18, 11], [11, 11]]]
        ),
    )


def test_set_obstacles(grid, obstacles):
    grid.set_obstacles(obstacles)

    expected_kcu = np.array(
        [102, 102, 102, 102, 98, 99, 103, 99, 99, 103], dtype=np.int32
    )
    assert_equal(grid.lines.kcu, expected_kcu)
    expected_flodu = np.array(
        [-0.1, -0.3, -0.3, -0.1, np.nan, np.nan, -0.1, np.nan, np.nan, -0.1],
        dtype=np.float64,
    )
    assert_almost_equal(grid.lines.flod, expected_flodu)
    assert_almost_equal(grid.lines.flou, expected_flodu)


def test_set_obstacles_no_intersect(grid, obstacles_no_intersect):
    grid.set_obstacles(obstacles_no_intersect)

    assert np.all(grid.lines.kcu != LineType.LINE_2D_OBSTACLE)
    assert_equal(grid.lines.flod, np.nan)
    assert_equal(grid.lines.flou, np.nan)


def test_compute_dpumax(lines: Lines, obstacles: Obstacles):
    expected = [-0.1, -0.3, -0.3, -0.1, np.nan, np.nan, -0.1, np.nan, np.nan, -0.1]
    actual, idx = obstacles.compute_dpumax(lines, where=np.arange(len(lines)))

    assert_equal(idx, [0, 1, 1, 0, -9999, -9999, 0, -9999, -9999, 0])
    assert_almost_equal(actual, expected)


def test_compute_dpumax_no_intersections(
    lines: Lines, obstacles_no_intersect: Obstacles
):
    actual, idx = obstacles_no_intersect.compute_dpumax(
        lines, where=np.arange(len(lines))
    )

    assert len(actual) == len(lines)
    assert_equal(idx, -9999)
    assert_almost_equal(actual, np.nan)


def test_compute_dpumax_no_obstacles(lines: Lines):
    actual, idx = Obstacles(id=[]).compute_dpumax(lines, where=np.arange(len(lines)))

    assert len(actual) == len(lines)
    assert_equal(idx, -9999)
    assert_almost_equal(actual, np.nan)


def test_compute_dpumax_where(lines: Lines, obstacles: Obstacles):
    expected = [-0.1, -0.1, np.nan, -0.1]
    actual, idx = obstacles.compute_dpumax(lines, where=[0, 3, 4, 6])

    assert_equal(idx, [0, 0, -9999, 0])
    assert_almost_equal(actual, expected)



import pytest

from threedigrid_builder.base import Nodes
from threedigrid_builder.constants import ContentType, NodeType


@pytest.fixture
def nodes():
    return Nodes(id=[1, 2, 3], coordinates=[(1, 10), (2, 12), (3, 13)])


def test_get_extent_1d(nodes):
    nodes = Nodes(
        id=range(3),
        coordinates=[(1, 10), (2, 12), (3, 13)],
        node_type=[
            NodeType.NODE_1D_STORAGE,
            NodeType.NODE_1D_NO_STORAGE,
            NodeType.NODE_2D_OPEN_WATER,
        ],
    )
    assert nodes.get_extent_1d() == (1, 10, 2, 12)


def test_get_extent_1d_from_1d_boundaries(nodes):
    nodes = Nodes(
        id=range(3),
        coordinates=[(1, 10), (2, 12), (3, 13)],
        node_type=[
            NodeType.NODE_1D_BOUNDARIES,
            NodeType.NODE_1D_BOUNDARIES,
            NodeType.NODE_1D_BOUNDARIES,
        ],
    )
    assert nodes.get_extent_1d() == (1, 10, 3, 13)


def test_get_extent_2d(nodes):
    nodes = Nodes(
        id=range(3),
        bounds=[
            (1, 10, 2, 11),
            (2, 12, 3, 13),
            (3, 13, 4, 14),
        ],
        node_type=[
            NodeType.NODE_2D_OPEN_WATER,
            NodeType.NODE_2D_OPEN_WATER,
            NodeType.NODE_1D_STORAGE,
        ],
    )
    assert nodes.get_extent_2d() == (1, 10, 3, 13)


def test_get_extent_none(nodes):
    nodes = Nodes(id=range(3))
    assert nodes.get_extent_1d() is None
    assert nodes.get_extent_2d() is None


CH = ContentType.TYPE_V2_CHANNEL
PI = ContentType.TYPE_V2_PIPE
CV = ContentType.TYPE_V2_CULVERT
CN = ContentType.TYPE_V2_CONNECTION_NODES


@pytest.mark.parametrize(
    "content_type,expected",
    [
        ([CH, CH, CN], "connection nodes [3], channels [1, 2]"),
        ([PI, CV], "pipes [1], culverts [2]"),
    ],
)
def test_format_message(content_type, expected):
    nodes = Nodes(
        id=range(len(content_type)),
        content_type=content_type,
        content_pk=range(1, len(content_type) + 1),
    )
    actual = nodes.format_message(nodes.id)

    assert actual == expected



import numpy as np
import pytest
from numpy.testing import assert_array_equal

from threedigrid_builder.grid import ExchangeLines


def test_channel_mapping():
    exchange_lines = ExchangeLines(id=range(1, 6), channel_id=[15, 2, 3, 15, 3])
    actual = exchange_lines.channel_mapping

    assert_array_equal(actual.id, [2, 3, 15])
    assert_array_equal(actual.exchange_line_id, [2, 3, 1])
    assert_array_equal(actual.secondary_exchange_line_id, [-9999, 5, 4])


@pytest.mark.parametrize(
    "channel_id,expected,is_primary",
    [
        (np.array([]), [], True),
        (np.array([1]), [-9999], True),
        (np.array([2]), [2], True),
        (np.array([2]), [-9999], False),
        (np.array([3]), [3], True),
        (np.array([3]), [5], False),
    ],
)
def test_get_for_channel_id(channel_id, expected, is_primary):
    exchange_lines = ExchangeLines(id=range(1, 6), channel_id=[15, 2, 3, 15, 3])
    actual = exchange_lines.get_for_channel_id(channel_id, is_primary)

    assert_array_equal(actual, expected)



import numpy as np
import shapely

from threedigrid_builder.base import search
from threedigrid_builder.constants import ContentType


def _compute_for_interpolated_nodes(nodes, cn_idx, objects):
    """Compute the initial waterlevels by interpolating between connection nodes

    This function is to be used for interpolated 1D nodes. It should be called
    after setting bottom levels (dmax) on connection nodes.

    Args:
        nodes (Nodes)
        cn_idx (ndarray of int): masks nodes so that you have only connection nodes
        objects (LinearObjects): channels, pipes, or culverts

    Returns:
        tuple of:
        - mask into nodes
        - initial waterlevel for each interpolated node in the mask
    """
    is_type = nodes.content_type == objects.content_type
    if not is_type.any():
        return is_type, None
    if shapely.is_missing(objects.the_geom).any():
        raise ValueError(
            f"{objects.__class__.__name__} found without a geometry. Call "
            f"set_geometries first."
        )
    lengths = shapely.length(objects.the_geom)

    idx = objects.id_to_index(nodes.content_pk[is_type])
    weights = nodes.s1d[is_type] / lengths[idx]
    if np.any(weights < 0.0) or np.any(weights > 1.0):
        raise ValueError("Encountered nodes outside of the linear object bounds")

    # convert connection_node_start_id to node indexes
    left_cn_idx = search(
        nodes.content_pk,
        objects.connection_node_start_id[idx],
        mask=cn_idx,
        assume_ordered=True,
    )
    right_cn_idx = search(
        nodes.content_pk,
        objects.connection_node_end_id[idx],
        mask=cn_idx,
        assume_ordered=True,
    )

    left = nodes.initial_waterlevel[left_cn_idx]
    right = nodes.initial_waterlevel[right_cn_idx]

    # Some logic to handle 'dry' initial waterlevels:
    # - (both sides nan): interpolated nodes become dry as well
    # - (one side nan): interpolate from initial_waterlevel to dmax
    # - (both value): interpolate between two initial_waterlevels
    has_value_left = np.isfinite(left)
    has_value_right = np.isfinite(right)
    left = np.where(
        ~has_value_left & has_value_right,
        nodes.dmax[left_cn_idx],
        left,
    )
    right = np.where(
        ~has_value_right & has_value_left,
        nodes.dmax[right_cn_idx],
        right,
    )

    return is_type, weights * right + (1 - weights) * left


def compute_initial_waterlevels(nodes, connection_nodes, channels, pipes, culverts):
    """Apply initial waterlevels (global or per connection nodes) to all 1D nodes.

    Bottom levels (dmax) should be set already.

    Args:
        connection_nodes (ConnectionNodes): used to map ids to indices
        channels (Channels)
        pipes (Pipes)
        culverts (Culverts)

    """
    if not np.any(np.isfinite(connection_nodes.initial_waterlevel)):
        return
    is_1d = nodes.content_type == ContentType.TYPE_V2_CONNECTION_NODES
    cn_idx = np.where(is_1d)[0]

    # First set the connection node initial waterlevels
    nodes.initial_waterlevel[cn_idx] = connection_nodes.initial_waterlevel[
        connection_nodes.id_to_index(nodes.content_pk[cn_idx])
    ]

    # Then interpolate
    for objects in (channels, pipes, culverts):
        is_type, initial_waterlevel = _compute_for_interpolated_nodes(
            nodes, cn_idx, objects
        )
        if initial_waterlevel is not None:
            nodes.initial_waterlevel[is_type] = initial_waterlevel
            is_1d |= is_type

    # Replace dry values with bottom levels (dmax)
    is_1d = np.where(is_1d)[0]
    is_dry = ~np.isfinite(nodes.initial_waterlevel[is_1d]) | (
        nodes.initial_waterlevel[is_1d] < nodes.dmax[is_1d]
    )
    nodes.initial_waterlevel[is_1d[is_dry]] = nodes.dmax[is_1d[is_dry]]



import numpy as np
import pytest
import shapely
from numpy.testing import assert_almost_equal, assert_equal

from threedigrid_builder.base import Lines, Nodes
from threedigrid_builder.constants import ContentType
from threedigrid_builder.grid import (
    Channels,
    CrossSectionDefinitions,
    CrossSectionLocations,
)
from threedigrid_builder.grid.cross_section_locations import (
    compute_bottom_level,
    compute_weights,
)


@pytest.fixture
def channels():
    return Channels(
        id=[51, 52, 53, 54],
        the_geom=[
            shapely.linestrings([(0, 0), (3, 0)]),  # 1 segment of size 3
            shapely.linestrings([(55, 3), (60, 3)]),  # 1 segment of size 5
            shapely.linestrings([(3, 1), (36, 1)]),  # 3 segments of size 11
            shapely.linestrings([(40, 2), (54, 2)]),  # 2 segments of size 7
        ],
    )


@pytest.fixture
def channel_nodes():
    CH = ContentType.TYPE_V2_CHANNEL
    CN = ContentType.TYPE_V2_CONNECTION_NODES
    return Nodes(
        id=range(8),
        content_type=[CN, CN, CN, CH, CH, CN, CH, CN],
        dmax=[0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8],
    )


@pytest.fixture
def channel_lines():
    # ordered by [content_pk, position on channel]
    return Lines(
        id=[0, 1, 2, 3, 4, 5, 6],
        content_type=ContentType.TYPE_V2_CHANNEL,
        content_pk=[51, 52, 53, 53, 53, 54, 54],
        ds1d=[3, 5, 11, 11, 11, 7, 7],
        s1d=[1.5, 2.5, 5.5, 16.5, 27.5, 3.5, 10.5],
        line=[(0, 1), (0, 2), (0, 3), (3, 4), (4, 5), (0, 6), (6, 7)],
        dpumax=[3.0, 3.0, 0, 0, 0, 0, 0],
    )


@pytest.fixture
def locations():
    return CrossSectionLocations(
        id=[1, 2, 3, 5, 6],
        the_geom=shapely.points([(1, 0), (26, 1), (16, 1), (58, 3), (49, 2)]),
        channel_id=[51, 53, 53, 52, 54],
        friction_type=[1, 1, 1, 2, 2],
        friction_value=[30, 35, 40, 0.02, 0.03],
        vegetation_stem_density=[1, 0.3, 0.4, np.nan, np.nan],
        vegetation_stem_diameter=[0.5, 0.5, 0.5, np.nan, np.nan],
        vegetation_height=[0.2, 1, 1, np.nan, np.nan],
        vegetation_drag_coefficient=[1, 1, 1, np.nan, np.nan],
        reference_level=[1.0, 2.0, 3.0, 5.0, 6.0],
        definition_id=[3, 3, 4, 4, 5],
    )


@pytest.fixture
def definitions():
    return CrossSectionDefinitions(id=[3, 4, 5])


@pytest.mark.parametrize("extrapolate", [True, False])
def test_compute_weights(locations, channels, channel_lines, extrapolate):
    """The test set has the following situation:

    - channel 51 with no added nodes and with 1 cs location
      O 1-X-- O
    - channel 52 with no added nodes and with 1 cs location:
      O --X-5 O
    - channel 53 with 2 added nodes, with 2 cs locations, extrapolation:
      O --X-- O 2-X-- O 3-X-- O
    - channel 54 with 1 added node, with 1 cs location, no extrapolation:
      O --X-- O 6-X-- O
    Here O means connection nodes, --X-- a line with a velocity point, and
    CS locations are given as numbers (ids).

    The 7 lines (velocity points) cover all possible situations:
    0. Only 1 CS location in the channel, extrapolation would lead to index -1
    1. Equalization on the right
    2. Extrapolation on the left
    3. Interpolation
    4. Extrapolation on the right
    5. Equalization on the left
    6. Only 1 CS location in the channel, extrapol. lead to too large index
    """
    expected_cross_loc1 = [1, 5, 3, 3, 3, 6, 6]
    expected_cross_loc2 = [1, 5, 2, 2, 2, 6, 6]
    expected_weight = [1.0, 1.0, 1.75, 0.65, -0.45, 1.0, 1.0]
    if not extrapolate:
        expected_weight = np.clip(expected_weight, 0, 1)

    cross_loc1, cross_loc2, cross_weights = compute_weights(
        channel_lines.content_pk, channel_lines.s1d, locations, channels, extrapolate
    )

    assert_equal(cross_loc1, expected_cross_loc1)
    assert_equal(cross_loc2, expected_cross_loc2)
    assert_almost_equal(cross_weights, expected_weight)


def test_compute_weights_edge_effects(channels):
    locations = CrossSectionLocations(
        id=range(3),
        the_geom=shapely.points([(0, 0), (55, 3), (3, 0)]),
        channel_id=[51, 52, 51],
    )
    expected_cross_loc1 = [0, 0, 1, 1]
    expected_cross_loc2 = [2, 2, 1, 1]
    expected_weight = [1.0, 0.0, 1.0, 1.0]

    cross_loc1, cross_loc2, cross_weights = compute_weights(
        [51, 51, 52, 52],
        [0.0, 3.0, 0.0, 55.0],
        locations,
        channels[:2],
        extrapolate=True,
    )

    assert_equal(cross_loc1, expected_cross_loc1)
    assert_equal(cross_loc2, expected_cross_loc2)
    assert_almost_equal(cross_weights, expected_weight)


@pytest.mark.parametrize(
    "extrapolate,expected",
    [
        (True, [1.0, 5.0, 3.75, 2.65, 1.55, 6.0, 6.0]),
        (False, [1.0, 5.0, 3.0, 2.65, 2.0, 6.0, 6.0]),
    ],
)
def test_compute_bottom_level(
    locations, channels, channel_lines, extrapolate, expected
):
    """Same setup as test_compute_weights, but now testing the derived bottom levels"""
    actual = compute_bottom_level(
        channel_lines.content_pk,
        channel_lines.s1d,
        locations,
        channels,
        extrapolate=extrapolate,
    )

    assert_almost_equal(actual, expected)


def test_apply_to_lines(channels, channel_lines, locations):
    locations.apply_to_lines(channel_lines, channels, extrapolate=False)

    assert_equal(channel_lines.cross_id1, [3, 4, 4, 4, 4, 5, 5])
    assert_equal(channel_lines.cross_id2, [3, 4, 3, 3, 3, 5, 5])
    assert_equal(channel_lines.frict_type1, [1, 2, 1, 1, 1, 2, 2])
    assert_equal(channel_lines.frict_type2, [1, 2, 1, 1, 1, 2, 2])
    assert_equal(channel_lines.frict_value1, [30.0, 0.02, 40.0, 40.0, 40.0, 0.03, 0.03])
    assert_equal(channel_lines.frict_value2, [30.0, 0.02, 35.0, 35.0, 35.0, 0.03, 0.03])
    assert_almost_equal(channel_lines.veg_coef1, [0.5, 0.0, 0.2, 0.2, 0.2, 0.0, 0.0])
    assert_almost_equal(channel_lines.veg_coef2, [0.5, 0.0, 0.15, 0.15, 0.15, 0.0, 0.0])
    assert_almost_equal(channel_lines.veg_height1, [0.2, 0.0, 1.0, 1.0, 1.0, 0.0, 0.0])
    assert_almost_equal(channel_lines.veg_height2, [0.2, 0.0, 1.0, 1.0, 1.0, 0.0, 0.0])
    assert_almost_equal(
        channel_lines.cross_weight, [1.0, 1.0, 1.0, 0.65, 0.0, 1.0, 1.0]
    )
    assert_almost_equal(
        channel_lines.invert_level_start_point, [1.0, 5.0, 3.0, 3.0, 2.1, 6.0, 6.0]
    )
    assert_almost_equal(
        channel_lines.invert_level_end_point, [1.0, 5.0, 3.0, 2.1, 2.0, 6.0, 6.0]
    )
    assert_almost_equal(channel_lines.dpumax, [1.0, 5.0, 3, 3.0, 2.1, 6.0, 6.0])


def test_apply_to_lines_extrapolate(channels, channel_lines, locations):
    locations.apply_to_lines(channel_lines, channels, extrapolate=True)

    assert_almost_equal(
        channel_lines.cross_weight, [1.0, 1.0, 1.75, 0.65, -0.45, 1.0, 1.0]
    )
    assert_almost_equal(
        channel_lines.invert_level_start_point, [1.0, 5.0, 4.3, 3.2, 2.1, 6.0, 6.0]
    )
    assert_almost_equal(
        channel_lines.invert_level_end_point, [1.0, 5.0, 3.2, 2.1, 1.0, 6.0, 6.0]
    )
    assert_almost_equal(channel_lines.dpumax, [1.0, 5.0, 4.3, 3.2, 2.1, 6.0, 6.0])


def test_apply_to_lines_project_locations(channels, channel_lines, locations):
    # CrossSectionLocations are not on the channels
    # - channel 51 (cs index 0): displaced 1 (should be projected on the line)
    # - channel 52 (cs index 3): displaced far away (should be projected to line end)
    locations.the_geom[0] = shapely.points((1, 1))  # id = 1
    locations.the_geom[3] = shapely.points((1000, 3))  # id = 5
    locations.apply_to_lines(channel_lines, channels, extrapolate=False)

    assert_equal(channel_lines.cross_id1, [3, 4, 4, 4, 4, 5, 5])
    assert_equal(channel_lines.cross_id2, [3, 4, 3, 3, 3, 5, 5])
    assert_almost_equal(
        channel_lines.cross_weight, [1.0, 1.0, 1.0, 0.65, 0.0, 1.0, 1.0]
    )



import numpy as np

from .array import Array

__all__ = ["Surfaces", "SurfaceMaps"]


class Surface:
    # Fields for every surface
    id: int
    code: str
    display_name: str
    function: str
    area: float
    centroid_x: float
    centroid_y: float
    dry_weather_flow: float
    nr_of_inhabitants: float
    infiltration_flag: bool
    outflow_delay: float
    storage_limit: float
    fb: float  # max_infiltration_capacity
    fe: float  # min_infiltration_capacity
    ka: float  # infiltration_decay_constant
    kh: float  # infiltration_recovery_constant

    # Only filled in for impervious surfaces
    surface_class: str
    surface_inclination: str
    surface_sub_class: str


class SurfaceMap:
    # Fields for mapping surface to connection node mapping
    # (surfaces can be connected to multiple connection_nodes)
    id: int
    fac: float  # Fraction of water on surface that is going to connection-node (0-1)
    imp: int  # 0-based index to surface array's (see Surface above)
    nxc: float  # Connection node x coordinate
    nyc: float  # Connection node y coordinate
    pk: int  # connection_node id
    cci: int  # node id


class Surfaces(Array[Surface]):
    """Zero-d surfaces."""

    def get_extent(self):
        if self.centroid_x is None or self.centroid_y is None:
            return None

        extent = (
            np.amin(self.centroid_x),
            np.amin(self.centroid_y),
            np.amax(self.centroid_x),
            np.amax(self.centroid_y),
        )
        if any(np.isnan(val) for val in extent):
            raise ValueError("Not all surface centroids have coordinates.")
        return extent


class SurfaceMaps(Array[SurfaceMap]):
    """Zero-d surfaces maps."""

    pass



from typing import Tuple

import numpy as np
import shapely

from threedigrid_builder.constants import CalculationType, ContentType, LineType

from .array import Array

__all__ = ["Lines"]


class Line:
    id: int
    code: str
    display_name: str
    kcu: LineType  # calculation type of the line
    line: Tuple[int, int]
    cross_pix_coords: Tuple[int, int, int, int]
    lik: int  # quadtree grid coordinate z of the line start
    lim: int  # quadtree grid coordinate x of the line start
    lin: int  # quadtree grid coordinate y of the line start
    s1d: float  # position (arclength) along a 1D element
    ds1d_half: float  # position (arclength) along a single line (between two calc nodes.)
    ds1d: float  # arclength
    line_geometries: shapely.Geometry
    line_coords: Tuple[float, float, float, float]
    content_type: ContentType
    content_pk: int
    dpumax: float  # bottom_level at the velocity point
    invert_level_start_point: float  # bottom level at line start
    invert_level_end_point: float  # bottom level at line end
    flod: float  # obstacle height
    flou: float  # obstacle height
    cross_id1: int  # the id of the cross section definition
    cross_id2: int  # the id of the cross section definition
    frict_type1: int
    frict_type2: int
    frict_value1: float  # For 1D-2Dgw: resistance factor-out (hydr. cond / thickness)
    frict_value2: float  # For 1D-2Dgw: resistance factor-in (hydr. cond / thickness)
    veg_coef1: float  # the product of vegetation properties (except height)
    veg_coef2: float  # the product of vegetation properties (except height)
    veg_height1: float  # the vegetaion height
    veg_height2: float  # the vegetaion height
    cross_weight: float  # For 1D-2Dgw: the exchange length (tables: cross_width)
    discharge_coefficient_positive: float
    discharge_coefficient_negative: float
    is_1d_boundary: int  # internal flag
    zoom_category: int
    connection_node_start_id: int
    connection_node_end_id: int
    crest_level: float
    crest_type: CalculationType
    dist_calc_points: float
    material: int
    sewerage: int
    sewerage_type: int
    windshieldings: Tuple[float, float, float, float, float, float, float, float]
    hydraulic_resistance_in: float
    hydraulic_resistance_out: float


class Lines(Array[Line]):
    """Line between two calculation nodes (a.k.a. velocity point)."""

    def get_velocity_points(self, mask):
        if np.any(shapely.is_missing(self.line_geometries[mask])):
            raise ValueError("No line geometries available")
        if np.any(~np.isfinite(self.ds1d_half[mask])):
            raise ValueError("No ds1d_half available")
        return shapely.line_interpolate_point(
            self.line_geometries[mask], self.ds1d_half[mask]
        )

    def set_discharge_coefficients(self):
        """Set discharge coefficients to 1.0 where unset."""
        for arr in (
            self.discharge_coefficient_positive,
            self.discharge_coefficient_negative,
        ):
            arr[np.isnan(arr)] = 1.0

    def set_line_coords(self, nodes):
        """Set line_coords from the node coordinates where necessary"""
        to_fix = np.isnan(self.line_coords).any(axis=1)
        start = nodes.coordinates[nodes.id_to_index(self.line[to_fix, 0])]
        end = nodes.coordinates[nodes.id_to_index(self.line[to_fix, 1])]
        self.line_coords[to_fix, :2] = start
        self.line_coords[to_fix, 2:] = end

    def fix_line_geometries(self):
        """Construct line_geometries from line_coords, where necessary"""
        to_fix = shapely.is_missing(self.line_geometries)
        if not to_fix.any():
            return
        if np.any(~np.isfinite(self.line_coords[to_fix])):
            raise ValueError("No line coords available")
        self.line_geometries[to_fix] = shapely.linestrings(
            self.line_coords[to_fix].reshape(-1, 2, 2)
        )

    def fix_ds1d(self):
        """Construct line distance (ds1d) from line_geometries, where necessary"""
        to_fix = np.isnan(self.ds1d)
        if not to_fix.any():
            return
        if np.any(shapely.is_missing(self.line_geometries[to_fix])):
            raise ValueError("No line geometries available")
        self.ds1d[to_fix] = shapely.length(self.line_geometries[to_fix])
        self.ds1d_half[to_fix] = 0.5 * self.ds1d[to_fix]

    def sort_by_nodes(self, node_ids):
        """Order selected lines by node id, ascending.

        Only lines with given node ids are ordered and then on then id that is supplied.
        This operation only makes sense if every node id occurs once in lines.line.
        """
        line_idx, start_or_end = np.where(np.isin(self.line, node_ids))
        new_line_idx = np.arange(len(self))
        sorter = np.argsort(self.line[line_idx, start_or_end])
        new_line_idx[line_idx] = new_line_idx[line_idx][sorter]
        self.reorder(new_line_idx)

    def set_2d_crest_levels(self, crest_levels, where):
        """Set obstacle crest levels to 2D lines

        Sets: flod, flou, kcu inplace (for 2D lines)
        """
        has_crest_level = where[np.isfinite(crest_levels)]
        is_2d_u = self.kcu[has_crest_level] == LineType.LINE_2D_U
        is_2d_v = self.kcu[has_crest_level] == LineType.LINE_2D_V
        self.kcu[has_crest_level[is_2d_u]] = LineType.LINE_2D_OBSTACLE_U
        self.kcu[has_crest_level[is_2d_v]] = LineType.LINE_2D_OBSTACLE_V
        self.flod[where] = crest_levels
        self.flou[where] = crest_levels



"""The interface layer (data conversion layer) of the project

A set of adapters that convert data from the external (HTTP, Database) to the internal
structure.

No code outside of this layer should contain details of a specific external data
structure.

The interface layer depends on the domain layer only.
"""

from .db import *  # NOQA
from .dict_out import *  # NOQA
from .geopackage import *  # NOQA
from .gridadmin import *  # NOQA
from .raster_gdal import *  # NOQA



import copy
from dataclasses import dataclass
from dataclasses import fields
from threedigrid_builder.base import is_int_enum
from threedigrid_builder.base import unpack_optional_type
from threedigrid_builder.constants import FrictionType
from threedigrid_builder.constants import InfiltrationSurfaceOption
from threedigrid_builder.constants import InitializationType
from threedigrid_builder.constants import InterflowType
from threedigrid_builder.exceptions import SchematisationError
from typing import Any, Dict, Optional

import numpy as np


__all__ = ["GridSettings", "TablesSettings"]


def greater_zero_check(obj, attr):
    value = getattr(obj, attr, None)
    if value is not None and value <= 0:
        raise SchematisationError(f"'{attr}' must be greater than 0.")


def replace_keys(dict: Dict[str, Any], key_map: Dict[str, str]) -> Dict[str, Any]:
    return {key if key not in key_map else key_map[key] : val for key, val in dict.items()}


@dataclass
class GridSettings:
    """Settings necessary for threedigrid-builder."""

    ## from GlobalSettings
    use_2d: bool
    use_1d_flow: bool
    use_2d_flow: bool
    use_0d_inflow: int
    grid_space: float
    dist_calc_points: float
    kmax: int
    embedded_cutoff_threshold: float = 0.05
    max_angle_1d_advection: float = 0.4 * np.pi

    @classmethod
    def from_dict(cls, dct):
        """Construct skipping unknown fields and None values"""
        schema_to_builder_map = {'minimum_cell_size': 'grid_space',
                                 'calculation_point_distance_1d': 'dist_calc_points',
                                 'nr_grid_levels': 'kmax'}
        dct = replace_keys(copy.copy(dct), schema_to_builder_map)
        class_fields = {f.name for f in fields(cls)}
        return cls(
            **{k: v for k, v in dct.items() if k in class_fields and v is not None}
        )

    def __post_init__(self):
        # validations
        if self.use_2d:
            for field in ["grid_sminimum_cell_sizepace", "nr_grid_levels"]:
                greater_zero_check(self, field)


@dataclass
class TablesSettings:
    """Settings necessary for threedi-tables."""

    ## from ModelSettings
    table_step_size: float
    frict_coef: float
    frict_coef_type: InitializationType
    frict_type: FrictionType = FrictionType.MANNING
    table_step_size_1d: float = None  # actual default is set in __post_init__
    maximum_table_step_size: float = None  # actual default  is set in __post_init__

    ## From Interception
    interception_global: Optional[float] = None
    interception_type: Optional[InitializationType] = None

    # TODO --> https://github.com/nens/threedigrid-builder/issues/86
    manhole_storage_area: Optional[float] = None

    # obstacle detection could be a tool in QGis?
    # dem_obstacle_detection: bool
    # dem_obstacle_height: float

    ## from Groundwater
    groundwater_impervious_layer_level: Optional[float] = None
    groundwater_impervious_layer_level_type: Optional[InitializationType] = None
    phreatic_storage_capacity: Optional[float] = None
    phreatic_storage_capacity_type: Optional[InitializationType] = None
    equilibrium_infiltration_rate: Optional[float] = None
    equilibrium_infiltration_rate_type: Optional[InitializationType] = None
    initial_infiltration_rate: Optional[float] = None
    initial_infiltration_rate_type: Optional[InitializationType] = None
    infiltration_decay_period: Optional[float] = None
    infiltration_decay_period_type: Optional[InitializationType] = None
    groundwater_hydro_connectivity: Optional[float] = None
    groundwater_hydro_connectivity_type: Optional[InitializationType] = None

    ## from Interflow
    interflow_type: InterflowType = InterflowType.NO_INTERLFOW
    porosity: Optional[float] = None
    porosity_type: Optional[InitializationType] = None
    porosity_layer_thickness: Optional[float] = None
    impervious_layer_elevation: Optional[float] = None
    hydraulic_conductivity: Optional[float] = None
    hydraulic_conductivity_type: Optional[InitializationType] = None

    ## from SimpleInfiltration
    infiltration_rate: Optional[float] = None
    infiltration_rate_type: Optional[InitializationType] = None
    infiltration_surface_option: InfiltrationSurfaceOption = (
        InfiltrationSurfaceOption.RAIN
    )
    max_infiltration_capacity: Optional[float] = None
    max_infiltration_capacity_type: Optional[InitializationType] = None

    ## from VegetationDrag
    vegetation_height: Optional[float] = None
    vegetation_height_type: Optional[InitializationType] = None
    vegetation_stem_count: Optional[float] = None
    vegetation_stem_count_type: Optional[InitializationType] = None
    vegetation_stem_diameter: Optional[float] = None
    vegetation_stem_diameter_type: Optional[InitializationType] = None
    vegetation_drag_coefficient: Optional[float] = None
    vegetation_drag_coefficient_type: Optional[InitializationType] = None

    def __post_init__(self):
        # defaults
        if self.table_step_size_1d is None:
            self.table_step_size_1d = self.table_step_size
        if self.maximum_table_step_size is None:
            self.maximum_table_step_size = 100 * self.table_step_size

        # validations
        for field in (
            "table_step_size",
            "table_step_size_1d",
            "maximum_table_step_size",
        ):
            greater_zero_check(self, field)

        if self.maximum_table_step_size < self.table_step_size:
            raise SchematisationError(
                f"'maximum_table_step_size' must not be less than 'table_step_size'."
            )

        # check enums
        for (name, elem_type) in self.__annotations__.items():
            elem_type = unpack_optional_type(elem_type)
            if is_int_enum(elem_type):
                value = getattr(self, name)
                if value is None:
                    continue
                try:
                    elem_type(value)
                except ValueError as e:
                    raise SchematisationError(str(e))

    @classmethod
    def from_dict(cls, dct):
        """Construct skipping unknown fields and None values"""
        class_fields = {f.name for f in fields(cls)}
        schema_to_builder_map = {"groundwater_hydraulic_conductivity": "groundwater_hydro_connectivity",
                                 "groundwater_hydraulic_conductivity_aggregation": "groundwater_hydro_connectivity_type",
                                 "groundwater_impervious_layer_level_aggregation": "groundwater_impervious_layer_level_type",
                                 "infiltration_decay_period_aggregation": "infiltration_decay_period_type",
                                 "initial_infiltration_rate_aggregation": "initial_infiltration_rate_type",
                                 "phreatic_storage_capacity_aggregation": "phreatic_storage_capacity_type",
                                 "max_infiltration_volume": "max_infiltration_capacity",
                                 "max_infiltration_volume_type": "max_infiltration_capacity_type",
                                 "manhole_aboveground_storage_area": "manhole_storage_area",
                                 "friction_coefficient": "frict_coef",
                                 "minimum_table_step_size": "table_step_size",
                                 "friction_type": "frict_type",
                                 "friction_coefficient_type": "frict_coef_type",
                                 "interception": "interception_global",
                                 }
        dct = replace_keys(copy.copy(dct), schema_to_builder_map)
        return cls(
            **{k: v for k, v in dct.items() if k in class_fields and v is not None}
        )



import numpy as np
import shapely

from threedigrid_builder.base import Array, search

__all__ = ["ExchangeLines"]


class ExchangeLine:
    id: int
    the_geom: shapely.Geometry
    channel_id: int
    exchange_level: float


class ExchangeLines(Array[ExchangeLine]):
    @property
    def channel_mapping(self) -> "ExchangeLineChannels":
        """Return 2 exchange lines ids for each channel id"""
        if not hasattr(self, "_channel_mapping"):
            channel_ids = np.unique(self.channel_id)
            result = np.full((len(channel_ids), 2), fill_value=-9999, dtype=np.int32)
            for i, mask in enumerate(self.split_in_two(self.channel_id)):
                result[:, i] = self.index_to_id(
                    search(
                        self.channel_id,
                        channel_ids,
                        mask=mask,
                        assume_ordered=False,
                        check_exists=False,
                    )
                )
            self._channel_mapping = ExchangeLineChannels(
                id=channel_ids,
                exchange_line_id=result[:, 0],
                secondary_exchange_line_id=result[:, 1],
            )
        return self._channel_mapping

    def get_for_channel_id(self, channel_id, is_primary):
        if len(self) == 0:
            return np.full_like(channel_id, fill_value=-9999)
        index = self.channel_mapping.id_to_index(channel_id)
        if is_primary:
            ids = self.channel_mapping.exchange_line_id
        else:
            ids = self.channel_mapping.secondary_exchange_line_id
        return np.where(index != -9999, np.take(ids, index, mode="clip"), -9999)


class ExchangeLineChannel:
    id: int
    exchange_line_id: int
    secondary_exchange_line_id: int


class ExchangeLineChannels(Array[ExchangeLineChannel]):
    pass



import itertools
import logging

import numpy as np
import shapely

from threedigrid_builder.base import Array, LineHalfs, Lines, Nodes, replace
from threedigrid_builder.constants import (
    CalculationType,
    ContentType,
    LineType,
    NodeType,
)

from .cross_section_locations import compute_bottom_level

logger = logging.getLogger(__name__)

__all__ = ["ConnectionNodes"]


class ConnectionNode:
    id: int
    the_geom: shapely.Geometry
    code: str
    storage_area: float
    manhole_id: int
    calculation_type: CalculationType
    manhole_indicator: int
    bottom_level: float
    drain_level: float
    surface_level: float
    shape: str  # enum with classes "00", "01", "02"
    width: float
    initial_waterlevel: float
    display_name: str
    zoom_category: int
    exchange_thickness: float
    hydraulic_conductivity_out: float
    hydraulic_conductivity_in: float


class ConnectionNodes(Array[ConnectionNode]):
    content_type = ContentType.TYPE_V2_CONNECTION_NODES

    def get_nodes(self, node_id_counter):
        """Convert connection nodes to a nodes instance

        Args:
            node_id_counter (iterable): an iterable yielding integers

        Returns:
            tuple of nodes (Nodes), segment_size (ndarray)
            nodes has data in the following columns:
            - id: 0-based counter generated here
            - coordinates
            - content_type: TYPE_V2_CHANNEL
            - content_pk: the id of the Channel from which this node originates
            - node_type: NODE_1D_STORAGE or NODE_1D_NO_STORAGE depending on storage_area
            - calculation_type: from calculation_type (which comes from manhole)
            - dmax: from bottom_level (which comes from manhole)
            - manhole_id: id of associated manhole.
            - drain_level: drain_level of associated manhole.
            - storage_area: area of connection_node.
        """
        node_type = np.where(
            self.storage_area >= 0,
            int(NodeType.NODE_1D_STORAGE),
            int(NodeType.NODE_1D_NO_STORAGE),
        )

        return Nodes(
            id=itertools.islice(node_id_counter, len(self)),
            coordinates=shapely.get_coordinates(self.the_geom),
            content_type=ContentType.TYPE_V2_CONNECTION_NODES,
            content_pk=self.id,
            node_type=node_type,
            calculation_type=self.calculation_type,
            dmax=self.bottom_level,
            manhole_id=self.manhole_id,
            drain_level=self.drain_level,
            storage_area=self.storage_area,
            display_name=self.display_name,
            zoom_category=self.zoom_category,
            manhole_indicator=self.manhole_indicator,
            has_groundwater_exchange=self.has_groundwater_exchange,
        )

    def is_closed(self, content_pk):
        """Whether object are 'closed' or 'open water' object.

        This is relevant for 1D-2D connections.
        """
        return self.storage_area[self.id_to_index(content_pk)] >= 0

    def get_1d2d_exchange_levels(self, content_pk, channels, locations, **kwargs):
        """Compute the exchange level (dpumax) for 1D-2D flowlines.

        For connection nodes 1D2D exchange levels are the drain levels. Where
        there is no drain level, the bank levels are taken from connected channels.

        Args:
            content_pk (array of int): object ids for which to compute levels
            channels (Channels): for the drain_levels
            locations (CrossSectionLocations): for the bank_levels

        Returns:
            exchange levels a.k.a. dpumax (array of float)
        """
        # get the corresponding connection_node ids and indexes
        connection_node_idx = self.id_to_index(content_pk)
        is_manhole = self.manhole_id[connection_node_idx] != -9999
        is_manhole_idx = connection_node_idx[is_manhole]

        # Check if manhole has drain_level below bottom_level
        has_lower_drn_lvl = (
            self.drain_level[is_manhole_idx] < self.bottom_level[is_manhole_idx]
        )
        if np.any(has_lower_drn_lvl):
            ids = self.manhole_id[is_manhole_idx[has_lower_drn_lvl]]
            logger.warning(
                f"Manholes {sorted(ids.tolist())} have a "
                f"bottom_level that is above drain_level."
            )

        # for open water nodes, compute drain level from channel bank levels
        open_water = ~is_manhole
        dpumax = np.full(len(self), np.nan)  # easier to initialize for all conn. nodes
        for name in ("connection_node_start_id", "connection_node_end_id"):
            has_node = np.isin(getattr(channels, name), content_pk[open_water])
            cn_idx_with_channel = self.id_to_index(getattr(channels, name)[has_node])
            if name == "connection_node_start_id":
                ds = 0.0
            else:
                ds = shapely.length(channels.the_geom[has_node])
            drain_level = compute_bottom_level(
                channels.id[has_node],
                ds,
                locations,
                channels,
                "bank_level",
            )
            _put_if_less(dpumax, cn_idx_with_channel, drain_level)
        # filter out connection nodes that were not in node_idx (non-connected ones)
        dpumax = dpumax[connection_node_idx]

        # for manholes: put in the drain level
        dpumax[is_manhole] = self.drain_level[is_manhole_idx]
        return dpumax

    @property
    def has_groundwater_exchange(self):
        with np.errstate(invalid="ignore"):
            return (
                (self.storage_area > 0)
                & (self.exchange_thickness > 0)
                & np.isfinite(self.hydraulic_conductivity_out)
                & np.isfinite(self.hydraulic_conductivity_in)
            )


def set_calculation_types(nodes: Nodes, lines: Lines):
    """Set the calculation types for connection nodes that do not yet have one.

    The calculation_type of a connection nodes is based on
      - connection_node.manhole.calculation_type (set in ConnectionNode.get_nodes)
      - if not present, the calculation_type of adjacent channels, pipes, and culverts.
        are taken. if these are different, the precedence is:
          EMBEDDED > ISOLATED > DOUBLE_CONNECTED > CONNECTED
      - if not present, then the calculation type becomes ISOLATED

    Args:
        nodes (Nodes): the nodes, those with content_type == TYPE_V2_CONNECTION_NODES
          and without a calculation_type will get a new calculation_type
        lines (Lines): the lines, including channels and pipes
    """
    PRIORITY = {
        LineType.LINE_1D_EMBEDDED: 0,
        LineType.LINE_1D_ISOLATED: 1,
        LineType.LINE_1D_DOUBLE_CONNECTED: 2,
        LineType.LINE_1D_CONNECTED: 3,
    }

    node_mask = (nodes.content_type == ContentType.TYPE_V2_CONNECTION_NODES) & (
        nodes.calculation_type == -9999
    )
    line_halfs = LineHalfs.for_connection_nodes(
        nodes,
        lines,
        line_mask=np.isin(
            lines.content_type,
            [
                ContentType.TYPE_V2_CHANNEL,
                ContentType.TYPE_V2_PIPE,
                ContentType.TYPE_V2_CULVERT,
            ],
        )
        & (lines.kcu != -9999),
        node_mask=node_mask,
    )
    line_halfs.reorder(
        np.lexsort([replace(line_halfs.kcu, PRIORITY), line_halfs.node_id])
    )

    calculation_type = line_halfs.first(line_halfs.kcu)

    # start off with ISOLATED
    nodes.calculation_type[node_mask] = CalculationType.ISOLATED
    # overwrite with the one derived from the line line_halfs
    nodes.calculation_type[
        nodes.id_to_index(line_halfs.get_reduce_id())
    ] = calculation_type


def _put_if_less(a, ind, v):
    """Replaces specified elements of an array with given values if  they are less."""
    # values can occur multiple times for the same index. sort descending by value so
    # that the lowest (latest) one will take precedence
    sorter = np.argsort(-v)
    v = v[sorter]
    ind = ind[sorter]
    # filter nan values
    is_nan = np.isnan(v)
    if is_nan.any():
        v = v[~is_nan]
        ind = ind[~is_nan]
    # same as v < a, except for how NaN is handled:
    is_less = ~(v > a[ind])
    if is_less.any():
        np.put(a, ind[is_less], v[is_less])


def set_bottom_levels(nodes: Nodes, lines: Lines):
    """Set the bottom level (dmax) for connection nodes that do not yet have one.

    The bottom level (dmax) of a connection nodes is based on:
      - from the manhole.bottom_level
        - must be lower than the invert level of pipes (if not: error)
      - if not present: the lowest of all neigboring objects. we use
        lines.invert_level_start_point to get levels of channels, pipes, culverts,
        weirs, and orifices.
        - (not implemented) if no storage: invert levels should not differ more than
          cross section height! -> check this in threedi-modelchecker

    Args:
        nodes (Nodes): the nodes, those with content_type == TYPE_V2_CONNECTION_NODES
          and without a dmax will get a new dmax
        lines (Lines): the lines, including channels, pipes, weirs, and culverts
    """
    # Compute the lowest invert level for each node connection node dmax will be the lowest of these object types:
    line_halfs = LineHalfs.for_connection_nodes(
        nodes,
        lines,
        line_mask=np.isin(
            lines.content_type,
            [
                ContentType.TYPE_V2_CHANNEL,
                ContentType.TYPE_V2_PIPE,
                ContentType.TYPE_V2_CULVERT,
                ContentType.TYPE_V2_WEIR,
                ContentType.TYPE_V2_ORIFICE,
            ],
        ),
    )
    line_halfs.reorder_by("node_id")
    dmax_per_node = line_halfs.nanmin(line_halfs.invert_level)
    node_idx = nodes.id_to_index(line_halfs.get_reduce_id())

    # The new dmax is the minimum of the existing one and the one from above computation
    dmax = np.fmin(nodes.dmax[node_idx], dmax_per_node)

    # Check if the new node dmax is below the original manhole dmax
    is_manhole = nodes.manhole_id[node_idx] != -9999
    has_lower_dmax = dmax[is_manhole] < nodes.dmax[node_idx[is_manhole]]
    if np.any(has_lower_dmax):
        ids = nodes.manhole_id[node_idx[is_manhole][has_lower_dmax]]
        logger.warning(
            f"Manholes {sorted(ids.tolist())} have a "
            f"bottom_level that is above one ore more of the following connected "
            f"objects: channel reference level, pipe/culvert invert level, "
            f"weir/orifice crest level."
        )

    # Place computed dmax values
    nodes.dmax[node_idx] = dmax



import itertools
from unittest import mock

import numpy as np
import pytest
import shapely
from numpy.testing import assert_almost_equal, assert_array_equal
from shapely.testing import assert_geometries_equal

from threedigrid_builder.base import Array, Lines, Nodes, PointsOnLine
from threedigrid_builder.constants import CalculationType, ContentType, NodeType
from threedigrid_builder.grid import ConnectionNodes, linear


@pytest.fixture(scope="session")
def random_lines():
    n_lines = 10  # approximate value
    n_coords = 100
    np.random.seed(0)
    indices = np.sort(np.random.randint(0, n_lines, n_coords))
    # discard lines with fewer than 2 coordinates
    indices = indices[~np.isin(indices, np.where(np.bincount(indices) < 2)[0])]
    coords = np.around(np.random.random((len(indices), 2)), decimals=6)
    lines = shapely.linestrings(coords, indices=indices)
    # discard nones
    lines = lines[~shapely.is_missing(lines)]
    return lines


@pytest.fixture
def two_lines():
    return np.array(
        [
            shapely.linestrings([(0, 10), (10, 10)]),
            shapely.linestrings([(0, 0), (6, 0), (6, 6)]),
        ]
    )


@pytest.fixture
def connection_nodes():
    # Used to map connection_node_start/end_id to an index (sequence id)
    return ConnectionNodes(
        id=np.array([21, 25, 33, 42]),
        the_geom=shapely.points([(0, 21), (1, 25), (2, 33), (3, 42)]),
        drain_level=[0.0, 20.0, 30.0, 10.0],
    )


class LinearObject:
    id: int
    the_geom: shapely.Geometry
    calculation_type: CalculationType
    connection_node_start_id: int
    connection_node_end_id: int
    dist_calc_points: float
    cross_section_definition_id: int
    invert_level_start_point: float
    invert_level_end_point: float
    discharge_coefficient_positive: float
    discharge_coefficient_negative: float
    display_name: str
    zoom_category: int
    material: int
    sewerage_type: int


class LinearObjects(Array[LinearObject], linear.BaseLinear):
    content_type = ContentType.TYPE_V2_1D_LATERAL  # just pick one for the test


@pytest.fixture
def one_linear_object():
    return LinearObjects(
        the_geom=shapely.linestrings([[(0, 0), (6, 0), (6, 6)]]),
        dist_calc_points=np.array([5.0]),
        id=np.array([1]),
        connection_node_start_id=np.array([21]),
        connection_node_end_id=np.array([42]),
        calculation_type=np.array([2]),
        cross_section_definition_id=[17],
    )


@pytest.fixture
def two_linear_objects():
    return LinearObjects(
        the_geom=shapely.linestrings(
            [[(0, 0), (10, 0), (10, 10)], [(0, 0), (0, 100), (100, 100)]]
        ),
        dist_calc_points=[5.0, np.nan],
        id=[1, 2],
        connection_node_start_id=[21, 25],
        connection_node_end_id=[42, 33],
        calculation_type=[2, 1],
        cross_section_definition_id=[17, 18],
        invert_level_start_point=[1.0, 2.0],
        invert_level_end_point=[3.0, 4.0],
        discharge_coefficient_positive=[0.8, 0.5],
        discharge_coefficient_negative=[0.5, 0.8],
        material=[-9999, 1],
        sewerage_type=[2, 1],
    )


@pytest.mark.parametrize(
    "global_dist_calc_points,expected_dists",
    [
        (74.0, [5.0, 74.0]),
        (0.0, [5.0, np.inf]),
        (-9999.0, [5.0, np.inf]),
        (np.nan, [5.0, np.inf]),
        (None, [5.0, np.inf]),
    ],
)
def test_interpolate_nodes(global_dist_calc_points, expected_dists, two_linear_objects):
    dummy_points = PointsOnLine(
        linestrings=two_linear_objects.linestrings,
        id=[0, 1, 2],
        s1d=[1.0, 2.0, 3.0],
        linestring_idx=[0, 0, 1],
    )
    with mock.patch.object(two_linear_objects.__class__, "linestrings") as linestrings:
        linestrings.segmentize().linestring_idx = [0, 1]
        linestrings.segmentize().interpolate_points.return_value = dummy_points
        nodes = two_linear_objects.interpolate_nodes(
            itertools.count(start=2), global_dist_calc_points
        )

    (fixed_nodes,), _ = linestrings.segmentize.call_args
    assert isinstance(fixed_nodes, PointsOnLine)
    assert len(fixed_nodes) == 0

    (dists,), _ = linestrings.segmentize().interpolate_points.call_args
    assert_almost_equal(dists, expected_dists)

    assert_array_equal(nodes.id, range(2, 2 + len(dummy_points)))
    assert_almost_equal(
        nodes.coordinates, shapely.get_coordinates(dummy_points.the_geom)
    )
    assert_array_equal(nodes.content_type, ContentType.TYPE_V2_1D_LATERAL)
    assert_array_equal(nodes.content_pk, [1, 1, 2])
    assert_array_equal(nodes.node_type, NodeType.NODE_1D_NO_STORAGE)
    assert_array_equal(nodes.calculation_type, [2, 2, 1])
    assert_array_equal(nodes.s1d, dummy_points.s1d)


def test_interpolate_nodes_skips_embedded(two_linear_objects):
    two_linear_objects.calculation_type[0] = CalculationType.EMBEDDED

    with mock.patch.object(LinearObjects, "linestrings") as linestrings:
        linestrings.segmentize().linestring_idx = [0, 1]
        linestrings.segmentize().interpolate_points.return_value = PointsOnLine.empty(
            two_linear_objects.linestrings
        )
        two_linear_objects.interpolate_nodes(
            itertools.count(start=2), global_dist_calc_points=74.0
        )

    (dists,), _ = linestrings.segmentize().interpolate_points.call_args
    assert_almost_equal(dists, [np.inf, 74.0])


def test_interpolate_nodes_with_fixture(two_linear_objects):
    fixed_nodes = PointsOnLine(
        linestrings=two_linear_objects.linestrings,
        id=[0, 1],
        s1d=[30.0, 50.0],
        linestring_idx=[1, 1],
        content_pk=[1, 2],
        secondary_content_pk=[-9999, 4],
    )
    dummy_points = PointsOnLine(
        linestrings=two_linear_objects.linestrings,
        id=[0, 1],
        s1d=[10.0, 125.0],
        linestring_idx=[0, 1],
    )
    with mock.patch.object(LinearObjects, "linestrings") as linestrings:
        linestrings.segmentize().linestring_idx = np.array([0, 1, 1, 1])
        linestrings.segmentize().interpolate_points.return_value = dummy_points
        nodes = two_linear_objects.interpolate_nodes(
            itertools.count(start=2),
            global_dist_calc_points=74.0,
            fixed_nodes=fixed_nodes,
        )

    (arg,), _ = linestrings.segmentize.call_args
    assert isinstance(arg, PointsOnLine)
    assert_array_equal(arg.s1d, fixed_nodes.s1d)

    (dists,), _ = linestrings.segmentize().interpolate_points.call_args
    assert_almost_equal(dists, [5.0, 74.0, 74.0, 74.0])

    assert_array_equal(nodes.id, [2, 3, 4, 5])
    assert_almost_equal(nodes.coordinates, [(10, 0), (0, 30), (0, 50), (25, 100)])
    assert_array_equal(nodes.content_type, ContentType.TYPE_V2_1D_LATERAL)
    assert_array_equal(nodes.content_pk, [1, 2, 2, 2])
    assert_array_equal(nodes.calculation_type, [2, 1, 1, 1])
    assert_array_equal(nodes.s1d, [10.0, 30.0, 50.0, 125.0])
    assert_array_equal(
        nodes.breach_ids, [(-9999, -9999), (1, -9999), (2, 4), (-9999, -9999)]
    )


def test_interpolate_nodes_with_fixture_on_start_end(one_linear_object):
    fixed_nodes = PointsOnLine(
        linestrings=one_linear_object.linestrings,
        id=[0, 1],
        s1d=[0.0, 12.0],
        linestring_idx=[0, 0],
        content_pk=[1, 2],
        secondary_content_pk=[-9999, 4],
    )
    dummy_points = PointsOnLine.empty(one_linear_object.linestrings)
    with mock.patch.object(LinearObjects, "linestrings") as linestrings:
        linestrings.segmentize().linestring_idx = np.array([0])
        linestrings.segmentize().interpolate_points.return_value = dummy_points
        nodes = one_linear_object.interpolate_nodes(
            itertools.count(start=2),
            global_dist_calc_points=74.0,
            fixed_nodes=fixed_nodes,
        )

    assert len(nodes) == 0


def test_get_lines(connection_nodes, two_linear_objects):
    nodes = Nodes(
        id=[10, 11, 12],
        content_pk=[1, 2, 2],
        s1d=[10.0, 50.0, 150.0],  # segments are not precisely the same size!
    )

    lines = two_linear_objects.get_lines(
        connection_nodes,
        nodes,
        itertools.count(start=0),
        connection_node_offset=100,
    )

    expected_line = [(100, 10), (10, 103), (101, 11), (11, 12), (12, 102)]
    expected_centers = [5.0, 15.0, 25.0, 100.0, 175.0]
    expected_sizes = [10.0, 10.0, 50.0, 100.0, 50.0]

    assert_array_equal(lines.id, range(5))
    assert_array_equal(lines.line, expected_line)
    assert_array_equal(lines.content_type, ContentType.TYPE_V2_1D_LATERAL)
    assert_array_equal(lines.content_pk, [1, 1, 2, 2, 2])
    assert_array_equal(lines.kcu, [2, 2, 1, 1, 1])
    assert_array_equal(lines.cross_id1, [17, 17, 18, 18, 18])
    assert_array_equal(lines.cross_id2, [17, 17, 18, 18, 18])
    assert_array_equal(lines.cross_weight, 1.0)
    assert_array_equal(lines.discharge_coefficient_positive, [0.8, 1.0, 0.5, 1.0, 1.0])
    assert_array_equal(lines.discharge_coefficient_negative, [1.0, 0.5, 1.0, 1.0, 0.8])
    assert_almost_equal(lines.s1d, expected_centers)
    assert_almost_equal(lines.ds1d, expected_sizes)
    assert_almost_equal(shapely.length(lines.line_geometries), expected_sizes)
    assert_almost_equal(lines.invert_level_start_point, [1.0, 2.0, 2.0, 2.5, 3.5])
    assert_almost_equal(lines.invert_level_end_point, [2.0, 3.0, 2.5, 3.5, 4])
    assert_almost_equal(lines.dpumax, [2, 3, 2.5, 3.5, 4])
    assert_almost_equal(lines.material, [-9999, -9999, 1, 1, 1])
    assert_almost_equal(lines.sewerage_type, [2, 2, 1, 1, 1])


def test_get_lines_embedded_mode(connection_nodes, two_linear_objects):
    nodes = Nodes(
        id=[10, 11, 12],
        content_pk=[1, 2, 2],
        s1d=[0.1] * 3,  # anything that isn't NaN
        embedded_in=[5, 4, 1],
    )

    two_linear_objects.calculation_type[:] = CalculationType.EMBEDDED
    lines = two_linear_objects.get_lines(
        connection_nodes,
        nodes,
        itertools.count(start=0),
        connection_node_offset=100,
        embedded_mode=True,
    )

    assert_array_equal(lines.line, [(100, 5), (5, 103), (101, 4), (4, 1), (1, 102)])


def test_get_lines_non_embedded_mode_skips(connection_nodes, two_linear_objects):
    # embedded linear objects are skipped
    two_linear_objects.calculation_type[1] = CalculationType.EMBEDDED
    lines = two_linear_objects.get_lines(
        connection_nodes,
        Nodes(id=[]),
        itertools.count(start=0),
        connection_node_offset=100,
        embedded_mode=False,
    )
    assert_array_equal(lines.line, [(100, 103)])


def test_get_lines_embedded_mode_skips(connection_nodes, two_linear_objects):
    # non-embedded linear objects are skipped
    two_linear_objects.calculation_type[1] = CalculationType.EMBEDDED
    lines = two_linear_objects.get_lines(
        connection_nodes,
        Nodes(id=[]),
        itertools.count(start=0),
        connection_node_offset=100,
        embedded_mode=True,
    )
    assert_array_equal(lines.line, [(101, 102)])


@pytest.mark.parametrize(
    "linear_object_idx,expected",
    [
        ([], [(0, 3)]),
        ([1], [(0, 4), (4, 3)]),
        ([1, 1, 1], [(0, 4), (4, 5), (5, 6), (6, 3)]),
    ],
)
def test_get_lines_one_linear_object(
    linear_object_idx, expected, connection_nodes, one_linear_object
):
    nodes = Nodes(
        id=range(4, 4 + len(linear_object_idx)),
        content_pk=linear_object_idx,
        s1d=[0.1] * (len(linear_object_idx)),  # some number, doesn't matter
    )
    lines = one_linear_object.get_lines(
        connection_nodes, nodes, itertools.count(start=0)
    )

    assert_array_equal(lines.line, expected)


@pytest.mark.parametrize(
    "linear_object_idx,expected",
    [
        ([], [(0, 3), (1, 2)]),
        ([1], [(0, 4), (4, 3), (1, 2)]),
        ([1, 1, 1], [(0, 4), (4, 5), (5, 6), (6, 3), (1, 2)]),
        ([2], [(0, 3), (1, 4), (4, 2)]),
        ([2, 2, 2], [(0, 3), (1, 4), (4, 5), (5, 6), (6, 2)]),
        ([1, 2, 2], [(0, 4), (4, 3), (1, 5), (5, 6), (6, 2)]),
        ([1, 1, 2], [(0, 4), (4, 5), (5, 3), (1, 6), (6, 2)]),
    ],
)
def test_get_lines_two_linear_objects(
    linear_object_idx, expected, connection_nodes, two_linear_objects
):
    nodes = Nodes(
        id=range(4, 4 + len(linear_object_idx)),
        content_pk=linear_object_idx,
        s1d=[0.1] * (len(linear_object_idx)),  # some number, doesn't matter
    )

    lines = two_linear_objects.get_lines(
        connection_nodes, nodes, itertools.count(start=0)
    )

    assert_array_equal(lines.line, expected)


@pytest.mark.parametrize(
    "geom,expected",
    [
        (None, [(0, 21), (3, 42)]),
        ([(0, 21), (3, 42)], [(0, 21), (3, 42)]),
        ([(0, 21), (1, 2), (3, 42)], [(0, 21), (1, 2), (3, 42)]),
        ([(0, 21), (3, 42)][::-1], [(0, 21), (3, 42)]),
        ([(0, 21), (1, 2), (3, 42)][::-1], [(0, 21), (1, 2), (3, 42)]),
        ([(1, 21), (3, 41)], [(0, 21), (1, 21), (3, 41), (3, 42)]),
        ([(3, 41), (1, 21)], [(0, 21), (1, 21), (3, 41), (3, 42)]),
    ],
)
def test_set_geometries(two_linear_objects, connection_nodes, geom, expected):
    two_linear_objects.the_geom[0] = shapely.linestrings(geom) if geom else None
    two_linear_objects.set_geometries(connection_nodes)

    expected_geometry = shapely.linestrings(expected)
    assert_geometries_equal(two_linear_objects.the_geom[0], expected_geometry)


def test_interpolate_nodes_no_geometries(two_linear_objects):
    two_linear_objects.the_geom[:] = None
    with pytest.raises(ValueError, match=".*encountered without a geometry."):
        two_linear_objects.interpolate_nodes(None, None)


@pytest.mark.parametrize(
    "ids,ds,expected",
    [
        ([1], [10.0], [5.0]),  # halfway the first object
        ([2], [100.0], [25.0]),  # halfway the second object
        ([1, 2], [10.0, 100.0], [5.0, 25.0]),
        ([2, 1], [100.0, 10.0], [25.0, 5.0]),
        ([1, 1, 1], [0.0, 5.0, 20.0], [0.0, 2.5, 10.0]),  # at 0%, 25%, 100%
        ([1, 2, 1], [0.0, 150.0, 20.0], [0.0, 27.5, 10.0]),  # mixed
    ],
)
def test_compute_drain_level(ids, ds, expected, two_linear_objects, connection_nodes):
    actual = two_linear_objects.compute_drain_level(ids, ds, connection_nodes)

    assert_almost_equal(actual, expected)


def test_compute_drain_level_with_one_nan(two_linear_objects, connection_nodes):
    connection_nodes.drain_level[:] = [0.0, np.nan, 1.0, np.nan]
    actual = two_linear_objects.compute_drain_level(
        [1, 2], [3.0, 4.0], connection_nodes
    )

    assert_almost_equal(actual, [0.0, 1.0])


def test_compute_drain_level_with_two_nan(two_linear_objects, connection_nodes):
    connection_nodes.drain_level[:] = [np.nan, 1.0, 1.0, np.nan]
    actual = two_linear_objects.compute_drain_level(
        [1, 2], [3.0, 4.0], connection_nodes
    )

    assert_almost_equal(actual, [np.nan, 1.0])


@mock.patch.object(LinearObjects, "has_groundwater_exchange", np.array([False, True]))
def test_apply_has_groundwater_exchange(two_linear_objects):
    nodes = Nodes(
        id=[1, 2, 3, 4],
    )
    lines = Lines(
        id=range(3),
        content_pk=[1, 2, 2],
        content_type=LinearObjects.content_type,
        line=[[2, 1], [1, 3], [3, 4]],
    )

    two_linear_objects.apply_has_groundwater_exchange(nodes, lines)

    assert_array_equal(nodes.has_groundwater_exchange, [1, 0, 1, 1])


@mock.patch.object(LinearObjects, "has_groundwater_exchange", np.array([False, True]))
def test_apply_has_groundwater_exchange_boundary_condition(two_linear_objects):
    nodes = Nodes(
        id=[1, 2, 3, 4],
        calculation_type=[
            CalculationType.CONNECTED,
            CalculationType.CONNECTED,
            CalculationType.CONNECTED,
            CalculationType.BOUNDARY_NODE,
        ],
    )
    lines = Lines(
        id=range(3),
        content_pk=[1, 2, 2],
        content_type=LinearObjects.content_type,
        line=[[2, 1], [1, 3], [3, 4]],
    )

    two_linear_objects.apply_has_groundwater_exchange(nodes, lines)

    assert_array_equal(nodes.has_groundwater_exchange, [1, 0, 1, 0])



import numpy as np
import pytest
import shapely
from numpy.testing import assert_array_equal

from threedigrid_builder.grid import GridRefinements

EPSILON = 1e-4


@pytest.fixture
def refinements():
    return GridRefinements(
        id=[5],
        refinement_level=[1],
        the_geom=[None],
    )


def test_rasterize_no_refinement():
    refinements = GridRefinements(id=[])
    actual = refinements.rasterize(
        origin=(0, 0), height=4, width=3, cell_size=0.5, no_data_value=99
    )
    assert actual.shape == (4, 3)
    assert actual.dtype == np.int32
    assert_array_equal(actual, 99)


def test_rasterize_none_refinement(refinements):
    actual = refinements.rasterize(
        origin=(0, 0), height=4, width=3, cell_size=0.5, no_data_value=99
    )
    assert actual.shape == (4, 3)
    assert actual.dtype == np.int32
    assert_array_equal(actual, 99)


def test_rasterize_outside(refinements):
    refinements.the_geom[0] = shapely.linestrings([[130.1, 11.9], [130.9, 11.9]])
    actual = refinements.rasterize(
        origin=(0, 0), height=4, width=3, cell_size=0.5, no_data_value=99
    )
    assert actual.shape == (4, 3)
    assert actual.dtype == np.int32
    assert_array_equal(actual, 99)


def test_rasterize_line_refinement(refinements):
    refinements.the_geom[0] = shapely.linestrings(
        [[15.0, 17.5], [17.5, 17.5], [17.5, 14.0]]
    )
    actual = refinements.rasterize(
        origin=(12.0, 10.0), height=8, width=6, cell_size=1.0, no_data_value=3
    )
    # Notes:
    # - origin == topleft (x, y order)
    # - geom starts at x=15.0, and does not touch cell (7, 2) which covers x=[14.0, 15.0)
    # - geom ends at y=14.0, and does not touch cell (3, 5) which covers x=[13.0, 14.0)
    expected = np.array(
        [
            [3, 3, 3, 3, 3, 3],
            [3, 3, 3, 3, 3, 3],
            [3, 3, 3, 3, 3, 3],
            [3, 3, 3, 3, 3, 3],
            [3, 3, 3, 3, 3, 1],
            [3, 3, 3, 3, 3, 1],
            [3, 3, 3, 3, 3, 1],
            [3, 3, 3, 1, 1, 1],
        ]
    )
    assert_array_equal(actual, expected)


def test_rasterize_line_refinement2(refinements):
    refinements.the_geom[0] = shapely.linestrings(
        [[15.0, 14.5], [12.5, 14.5], [12.5, 16.0]]
    )
    actual = refinements.rasterize(
        origin=(12.0, 10.0), height=8, width=6, cell_size=1.0, no_data_value=3
    )
    # Notes:
    # - origin == topleft (x, y order)
    # - geom starts at x=15.0, and does touch cell (4, 3) which covers x=[15.0, 16.0)
    # - geom ends at y=16.0, and does touch cell (6, 0) which covers y=[16.0, 17.0)
    expected = np.array(
        [
            [3, 3, 3, 3, 3, 3],
            [3, 3, 3, 3, 3, 3],
            [3, 3, 3, 3, 3, 3],
            [3, 3, 3, 3, 3, 3],
            [1, 1, 1, 1, 3, 3],
            [1, 3, 3, 3, 3, 3],
            [1, 3, 3, 3, 3, 3],
            [3, 3, 3, 3, 3, 3],
        ]
    )
    assert_array_equal(actual, expected)


def test_rasterize_poly_refinement(refinements):
    refinements.the_geom[0] = shapely.box(
        13.0 + EPSILON, 11.0 + EPSILON, 14.0 + EPSILON, 13.0 + EPSILON
    )
    actual = refinements.rasterize(
        origin=(12.0, 10.0), height=8, width=6, cell_size=0.5, no_data_value=3
    )
    # Notes:
    # - origin == topleft (x, y order)
    # - cell_size == 0.5 in this test
    # - geom covers x=[13.0, 14.0], so it excludes column 1 [12.5, 13.0)
    #   and includes column 4 [14.0, 14.5)
    # - geom covers y=[11.0, 13.0] so it excludes row 1 [10.5, 11.0) and
    #   includes row 6 [13.0, 13.5)
    expected = np.array(
        [
            [3, 3, 3, 3, 3, 3],
            [3, 3, 3, 3, 3, 3],
            [3, 3, 1, 1, 1, 3],
            [3, 3, 1, 1, 1, 3],
            [3, 3, 1, 1, 1, 3],
            [3, 3, 1, 1, 1, 3],
            [3, 3, 1, 1, 1, 3],
            [3, 3, 3, 3, 3, 3],
        ]
    )
    assert_array_equal(actual, expected)


def test_quadtree_small_line_refinement(refinements):
    refinements.the_geom[0] = shapely.linestrings([[13.1, 11.9], [13.9, 11.9]])
    actual = refinements.rasterize(
        origin=(12.0, 10.0), height=8, width=6, cell_size=1.0, no_data_value=3
    )
    expected = np.array(
        [
            [3, 3, 3, 3, 3, 3],
            [3, 1, 3, 3, 3, 3],
            [3, 3, 3, 3, 3, 3],
            [3, 3, 3, 3, 3, 3],
            [3, 3, 3, 3, 3, 3],
            [3, 3, 3, 3, 3, 3],
            [3, 3, 3, 3, 3, 3],
            [3, 3, 3, 3, 3, 3],
        ]
    )
    assert_array_equal(actual, expected)


@pytest.mark.parametrize("reverse", [False, True])
def test_rasterize_multiple_refinements(reverse):
    refinements = GridRefinements(
        id=[5, 8],
        refinement_level=[1, 2],
        the_geom=[
            shapely.box(13.0 + EPSILON, 11.0 + EPSILON, 14.0 + EPSILON, 13.0 + EPSILON),
            shapely.box(12.5 + EPSILON, 10.5 + EPSILON, 13.5 + EPSILON, 12.5 + EPSILON),
        ],
    )
    if reverse:
        refinements.refinement_level = refinements.refinement_level[::-1]
        refinements.the_geom = refinements.the_geom[::-1]
    actual = refinements.rasterize(
        origin=(12.0, 10.0), height=8, width=6, cell_size=0.5, no_data_value=3
    )
    # Notes:
    # - see test_rasterize_poly_refinement
    # - lowest refinement_level precedes
    expected = np.array(
        [
            [3, 3, 3, 3, 3, 3],
            [3, 2, 2, 2, 3, 3],
            [3, 2, 1, 1, 1, 3],
            [3, 2, 1, 1, 1, 3],
            [3, 2, 1, 1, 1, 3],
            [3, 2, 1, 1, 1, 3],
            [3, 3, 1, 1, 1, 3],
            [3, 3, 3, 3, 3, 3],
        ]
    )
    assert_array_equal(actual, expected)



from unittest import mock

import pytest
import shapely
from numpy import nan
from numpy.testing import assert_almost_equal

from threedigrid_builder.base import Nodes
from threedigrid_builder.constants import ContentType
from threedigrid_builder.grid import Channels, ConnectionNodes
from threedigrid_builder.grid.initial_waterlevels import compute_initial_waterlevels


@pytest.fixture
def channels():
    return Channels(
        id=[51, 52, 53, 54],
        connection_node_start_id=[1, 1, 1, 1],
        connection_node_end_id=[2, 3, 4, 5],
        the_geom=[
            shapely.linestrings([(0, 0), (3, 0)]),  # 1 segment of size 3
            shapely.linestrings([(55, 3), (60, 3)]),  # 1 segment of size 5
            shapely.linestrings([(3, 1), (36, 1)]),  # 3 segments of size 11
            shapely.linestrings([(40, 2), (54, 2)]),  # 2 segments of size 7
        ],
    )


@pytest.fixture
def nodes():
    CH = ContentType.TYPE_V2_CHANNEL
    CN = ContentType.TYPE_V2_CONNECTION_NODES
    return Nodes(
        id=range(8),
        content_pk=[1, 2, 3, 53, 53, 4, 54, 5],
        content_type=[CN, CN, CN, CH, CH, CN, CH, CN],
        s1d=[nan, nan, nan, 11.0, 22.0, nan, 7.0, nan],
        dmax=[0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8],
    )


@pytest.fixture
def connection_nodes():
    return ConnectionNodes(id=[1, 2, 3, 4, 5])


@pytest.mark.parametrize(
    "initial_waterlevels, expected",
    [
        ([0.0, 0.0, 0.0, 0.0, 0.0], [0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8]),
        ([1.0, 1.0, 1.0, 1.0, 1.0], [1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0]),
        ([1.0, nan, nan, nan, nan], [1.0, 0.2, 0.3, 0.867, 0.733, 0.6, 0.9, 0.8]),
        ([nan, nan, nan, nan, 1.0], [0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 1.0]),
        ([nan, nan, nan, nan, 2.0], [0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 1.05, 2.0]),
    ],
)
def test_compute_initial_waterlevels(
    nodes, channels, connection_nodes, initial_waterlevels, expected
):
    connection_nodes.initial_waterlevel[:] = initial_waterlevels
    compute_initial_waterlevels(
        nodes,
        connection_nodes=connection_nodes,
        channels=channels,
        pipes=mock.Mock(),
        culverts=mock.Mock(),
    )

    assert_almost_equal(nodes.initial_waterlevel, expected, decimal=3)



import typing
from enum import IntEnum
from typing import _GenericAlias, Generic, TypeVar

import numpy as np

__all__ = [
    "is_tuple_type",
    "is_int_enum",
    "unpack_optional_type",
    "replace",
    "search",
    "Array",
    "TooManyExist",
]


def is_tuple_type(_type):
    """Test if the type is a indexed typing.Tuple type.

    Examples:
        is_tuple_type(Tuple[int, int]) == True
    """
    return isinstance(_type, _GenericAlias) and _type.__origin__ is tuple


def is_union_type(_type):
    """Test if the type is an typing.Union type.

    Examples:
        is_union_type(Union[int, str]) == True
    """
    return isinstance(_type, _GenericAlias) and _type.__origin__ is typing.Union


def is_int_enum(_type):
    """Test if the type is a subclass of IntEnum."""
    try:
        return issubclass(_type, IntEnum)
    except TypeError:  # happens when _type is something like List[int]
        return False


def unpack_optional_type(_type):
    """Unpack Optional[<type>] to <type> if necessary

    Examples:
        >>> unpack_optional(Optional[int])
        int
        >>> unpack_optional(int)
        int
    """
    # Optional[x] is actually Union[x, None]
    if not is_union_type(_type):
        return _type
    # Check if it has 2 args and the second one is a NoneType
    if len(_type.__args__) == 2 and isinstance(None, _type.__args__[1]):
        return _type.__args__[0]


def _get_dtype_and_null_value(elem_type):
    if elem_type is int or is_int_enum(elem_type):
        dtype = np.int32
        null_value = -9999
    elif elem_type is float:
        dtype = np.float64
        null_value = np.nan
    elif elem_type is bool:
        dtype = bool
        null_value = False
    else:
        dtype = object
        null_value = None
    return dtype, null_value


def _to_ndarray(value, elem_type, expected_length):
    """Cast value to numpy array, with some type mappings.

    Args:
        value (iterable or None): iterable to convert to ndarray
        elem_type (type)
        expected_length (int or None): the expected number of rows
    """
    # Tuple[...] becomes 2D Array
    if is_tuple_type(elem_type):
        sub_types = elem_type.__args__
        expected_shape = (expected_length, len(sub_types))
        elem_type = sub_types[0]
    else:
        expected_shape = (expected_length,)

    # cast python to numpy dtypes
    dtype, null_value = _get_dtype_and_null_value(elem_type)

    # return empty array if no value or None is supplied
    if value is None:
        return np.full(expected_shape, null_value, dtype=dtype, order="F")
    elif np.isscalar(value) or isinstance(value, IntEnum):
        arr = np.full(expected_shape, value, dtype=dtype, order="F")
    else:
        arr = np.asarray(value, dtype=dtype, order="F")

    if is_int_enum(elem_type):
        allowed = list(elem_type) + [-9999]
        if not np.isin(arr, allowed).all():
            raise ValueError(
                f"{set(value) - set(allowed)} is not a valid {elem_type.__name__}"
            )
    if arr.shape != expected_shape:
        raise ValueError(
            f"Expected an array with shape {expected_shape}, got {arr.shape}."
        )
    return arr


T = TypeVar("T")


class Array(Generic[T]):
    """A dataclass with fields ("columns") that are numpy arrays of equal size."""

    scalars = tuple()

    def __init_subclass__(cls) -> None:
        # Trick to get the record dataclass from the generic typing
        base = cls.__orig_bases__[0]  # type: ignore
        (data_class,) = base.__args__
        cls.validate_data_class(data_class)
        super().__init_subclass__()
        cls.data_class = data_class

    @staticmethod
    def validate_data_class(data_class):
        fields = typing.get_type_hints(data_class)

        # id must be present
        if "id" not in fields:
            raise TypeError("The id field is required")
        if fields["id"] is not int:
            raise TypeError("The id field should be of integer type")

        # check the subtypes of Tuples
        for name, elem_type in fields.items():
            if is_tuple_type(elem_type):
                sub_types = elem_type.__args__
                n_columns = len(sub_types)
                if n_columns == 0:
                    raise TypeError("Tuple cannot contain 0 elements")
                elem_type = sub_types[0]
                if not all(x is elem_type for x in sub_types):
                    raise TypeError("Tuple cannot contain mixed dtypes")

    def __init__(self, **kwargs):
        """Initialize the array dataclass from values.

        None values are converted to arrays with NULL values that are appropriate
        for the type (nan for float, -9999 for int, None for object,
        False for bool).
        """
        # id is required; get its length to use it for validation
        id = kwargs.pop("id", None)
        if id is None:
            raise TypeError("missing required keyword argument 'id'")
        # handle generators
        if hasattr(id, "__iter__") and not hasattr(id, "__len__"):
            id = np.fromiter(id, dtype=np.int32)
        # cast to integer array
        id = np.asarray(id, dtype=np.int32, order="F")
        if id.ndim != 1:
            raise ValueError(f"Expected one-dimensional 'id' array, got {id.ndim}")
        if id.size > 1 and not np.all(id[1:] > id[:-1]):
            raise ValueError("Values in 'id' must be unique and sorted")
        self.id = id

        # fetch the scalar arguments
        for name in self.scalars:
            value = kwargs.pop(name, None)
            if value is None:
                raise TypeError(f"missing required keyword argument '{name}'")
            setattr(self, name, value)

        # convert each field to an array and set it to self
        fields = typing.get_type_hints(self.data_class)
        for name, elem_type in fields.items():
            if name == "id":
                continue
            value = kwargs.pop(name, None)
            if value is not None:
                try:
                    arr = _to_ndarray(value, elem_type, id.size)
                except ValueError as e:
                    raise ValueError(f"Error parsing {name} to array: {e}")
                setattr(self, name, arr)

        # call the init of the wrapped class
        if len(kwargs) > 0:
            raise TypeError(
                f"{self.__class__.__name__} got unexpected arguments {list(kwargs)}"
            )
        super().__init__(**kwargs)

    def __repr__(self):
        return (
            f"<{self.__class__.__name__} object, "
            f"{self.data_class.__name__} array of length {len(self)}>"
        )

    def __len__(self):
        return len(self.id)

    def __add__(self, other: T) -> T:
        """Concatenate two array dataclasses of equal type."""
        if not issubclass(other.__class__, self.__class__):
            raise TypeError(
                f"Cannot concatenate {self} with {other} as they are not of "
                f"equal types."
            )
        new_fields = {
            name: np.concatenate((getattr(self, name), getattr(other, name)))
            for name in typing.get_type_hints(self.data_class).keys()
        }
        for scalar_arg in self.scalars:
            new_fields[scalar_arg] = getattr(self, scalar_arg)
        return self.__class__(**new_fields)

    def id_to_index(self, id, check_exists: bool = False):
        """Find the index of records with given id.

        Args:
            id (int or array_like): The id(s) to find in self.id
            check_exists (bool): Whether to check if the id is actually present. Raises
              a KeyError with added "values" and "indices" attributes if an id is not
              present.

        Returns:
            int or array_like: the indexes into self.id
        """
        # some timings:
        # len(id)    len(self.id)   timing (microseconds)
        # 1000       2000           44
        # 1000       10000          62
        return search(self.id, id, assume_ordered=True, check_exists=check_exists)

    def index_to_id(self, index):
        """Find the id of records with given index.

        Note:
            Index should be 0 <= index < len(self). Values of -9999 will persist.

        Args:
            index (int or array_like): The indices to return from self.id

        Returns:
            int or array_like: the ids from self.id
        """
        index = np.asarray(index)
        return self.take("id", index)

    def __getattr__(self, name):
        try:
            elem_type = typing.get_type_hints(self.data_class)[name]
        except KeyError:
            raise AttributeError
        arr = _to_ndarray(None, elem_type, len(self))
        setattr(self, name, arr)
        return arr

    def take(self, name, index, fill_value=None):
        """Take values from a column, returing 'fill_value' where index == -9999.

        Args:
            name (str): the column name to take values from
            index (int or array_like): The indices to return from self.id
            fill_value: what to return for -9999 indices (defaults to -9999/NaN/None)

        Returns:
            scalar or array_like: the values from column 'name'
        """
        values = getattr(self, name)
        if fill_value is None:
            _, fill_value = _get_dtype_and_null_value(
                typing.get_type_hints(self.data_class)[name]
            )
        index = np.asarray(index)
        return np.where(index != -9999, np.take(values, index, mode="clip"), fill_value)

    def to_dict(self):
        fields = typing.get_type_hints(self.data_class)
        return {field: getattr(self, field) for field in fields}

    def __getitem__(self, idx) -> T:
        """Create a masked copy of this array dataclass"""
        args = {}
        for field in typing.get_type_hints(self.data_class):
            args[field] = getattr(self, field)[idx]
        for scalar_arg in self.scalars:
            args[scalar_arg] = getattr(self, scalar_arg)
        return self.__class__(**args)

    def reorder(self, idx) -> None:
        """Reorder self by given index, inplace.

        Note that this skips self.id: the records are renumbered.
        """
        for field in typing.get_type_hints(self.data_class):
            if field == "id":
                continue
            setattr(self, field, getattr(self, field)[idx])

    def reorder_by(self, attr: str, **kwargs) -> None:
        """Reorder self by given column, inplace.

        Note that this skips self.id: the records are renumbered.
        """
        idx = np.argsort(getattr(self, attr), **kwargs)
        self.reorder(idx)

    def split_in_two(self, by):
        """Split self in to 2 arrays having the first and second occurence of 'by'.

        Returned are index arrays that index into self.

        If 'by' contains only unique values, the second array will be empty. If any value
        occurs more than 2 times, a 'TooManyExist' error is raised, having a 'values'
        attribute of the values that occur more than 2 times.
        """
        if len(by) != len(self):
            raise ValueError("'by' must have the same length as 'self'")
        values, idx1, counts = np.unique(by, return_index=True, return_counts=True)
        if np.any(counts > 2):
            raise TooManyExist("", values=values[counts > 2])
        assert not np.any(counts > 2)
        idx2 = np.delete(np.arange(len(self)), idx1)
        return idx1, idx2


class TooManyExist(KeyError):
    def __init__(self, msg, values):
        self.values = values
        super().__init__(msg)


def replace(arr, mapping, check_present=False):
    """Return array with its values replaced according to ``mapping``.

    If ``check_present`` is False, it is assumed that all elements in the array are keys
    of ``mapping``.
    """
    keys, values = np.array(sorted(mapping.items())).T
    indices = np.digitize(arr, keys, right=True)
    if check_present and not np.all(keys[indices] == arr):
        raise ValueError("Not all values are present in the replacement dict")
    return values[indices]


class DoesNotExist(KeyError):
    def __init__(self, msg, values, indices):
        self.values = values
        self.indices = indices
        super().__init__(msg)


def search(a, v, mask=None, assume_ordered=False, check_exists=True):
    """Find indices where `v` occurs in `a`.

    Args:
      a (1D array_like): Input array. If assume_ordered is True, then it must be sorted
        in ascending order.
      v (array_like): Values to find in a.
      mask (1D array_like): A (boolean or index) mask to apply to a before searching.
      assume_ordered (bool, optional): Whether to assume a is ordered, default False.
      check_exists (bool, optional): Check whether the values in v exist in a,
        default True.

    Returns:
      1D array with the same shape as `v`

    Raises:
      If check_exists is True or if the (masked) ``a`` is empty, this function raises a
      KeyError with added ``values`` and ``indices`` attributes corresponding to the
      missing values and the indices of them into ``v``.
    """
    v = np.asarray(v)
    if v.size == 0:
        return np.empty_like(v, dtype=int)

    if mask is not None:
        mask = np.asarray(mask)
        if mask.dtype == bool:
            mask = np.where(mask)[0]
        a = np.take(a, mask)

    # If there is no array to search in: raise directly
    if len(a) == 0:
        if check_exists:
            raise DoesNotExist(
                "search encountered missing elements",
                values=v,
                indices=np.arange(v.shape[0]),
            )
        return np.full(len(v), -9999, dtype=int)

    if assume_ordered:
        ind = np.searchsorted(a, v)
    else:
        sorter = np.argsort(a)
        ind = np.take(sorter, np.searchsorted(a, v, sorter=sorter), mode="clip")

    missing = np.take(a, ind, mode="clip") != v
    if check_exists and missing.any():
        raise DoesNotExist(
            "search encountered missing elements",
            values=np.compress(missing, v),
            indices=np.where(missing)[0],
        )
    elif missing.any():
        if np.isscalar(ind):
            ind = -9999
        else:
            ind[missing] = -9999

    # Map to original (unmasked) indices
    if mask is not None:
        ind = np.where(ind != -9999, np.take(mask, ind, mode="clip"), -9999)

    return ind



import itertools

import numpy as np

from threedigrid_builder.base import Lines, Nodes
from threedigrid_builder.constants import LineType, NodeType
from threedigrid_builder.exceptions import SchematisationError


def get_nodes(nodes, node_id_counter):
    """Add groundwater nodes from open water nodes."""
    open_water_nodes = nodes[np.isin(nodes.node_type, NodeType.NODE_2D_OPEN_WATER)]
    if len(open_water_nodes) == 0:
        raise SchematisationError(
            "Unable to create groundwater nodes, no open water nodes found."
        )
    id_n = itertools.islice(node_id_counter, len(open_water_nodes))
    groundwater_nodes = Nodes(
        id=id_n,
        node_type=NodeType.NODE_2D_GROUNDWATER,
        bounds=open_water_nodes.bounds,
        pixel_coords=open_water_nodes.pixel_coords,
        nodk=open_water_nodes.nodk,
        nodm=open_water_nodes.nodm,
        nodn=open_water_nodes.nodn,
        coordinates=open_water_nodes.coordinates,
    )
    node_id_offset = groundwater_nodes.id[0] - open_water_nodes.id[0]
    return groundwater_nodes, node_id_offset


def get_vertical_lines(nodes, node_id_offset, line_id_counter):
    """Add vertical lines between open water - and groundwater nodes."""
    open_water_nodes = nodes[np.isin(nodes.node_type, NodeType.NODE_2D_OPEN_WATER)]

    id_n = itertools.islice(line_id_counter, len(open_water_nodes))
    return Lines(
        id=id_n,
        kcu=LineType.LINE_2D_VERTICAL,
        line=np.stack(
            (open_water_nodes.id, open_water_nodes.id + node_id_offset), axis=1
        ),
        lik=open_water_nodes.nodk,
        lim=open_water_nodes.nodm,
        lin=open_water_nodes.nodn,
    )


def get_lines(lines, node_id_offset, line_id_counter):
    """Add groundwater lines from open water - and obstacle lines."""
    open_water_lines = lines[
        np.isin(
            lines.kcu,
            (
                LineType.LINE_2D_U,
                LineType.LINE_2D_V,
                LineType.LINE_2D_OBSTACLE_U,
                LineType.LINE_2D_OBSTACLE_V,
            ),
        )
    ]
    id_n = itertools.islice(line_id_counter, len(open_water_lines))
    return Lines(
        id=id_n,
        kcu=LineType.LINE_2D_GROUNDWATER,
        line=open_water_lines.line + node_id_offset,
        lik=open_water_lines.lik,
        lim=open_water_lines.lim,
        lin=open_water_lines.lin,
        line_coords=open_water_lines.line_coords,
        cross_pix_coords=open_water_lines.cross_pix_coords,
    )
